"""Model definitions."""

import torch
import torchvision
from torchvision.models.resnet import BasicBlock, Bottleneck

from collections import OrderedDict

from .mobilenet import MobileNetV2
from .vgg import VGG


def get_model(model_name, dataset_name, pretrained=False):
    """Retrieve an appropriate architecture."""
    if 'CIFAR' in dataset_name or 'MNIST' in dataset_name:
        if pretrained:
            raise ValueError('Loading pretrained models is only supported for ImageNet.')
        in_channels = 1 if dataset_name == 'MNIST' else 3
        num_classes = 10 if dataset_name in ['CIFAR10', 'MNIST'] else 100
        if 'ResNet' in model_name:
            model = resnet_picker(model_name, dataset_name)
        elif 'efficientnet-b' in model_name.lower():
            from efficientnet_pytorch import EfficientNet
            model = EfficientNet.from_name(model_name.lower())
        elif model_name == 'ConvNet':
            model = convnet(width=32, in_channels=in_channels, num_classes=num_classes)
        elif model_name == 'ConvNet64':
            model = convnet(width=64, in_channels=in_channels, num_classes=num_classes)
        elif model_name == 'ConvNet128':
            model = convnet(width=64, in_channels=in_channels, num_classes=num_classes)
        elif model_name == 'ConvNetBN':
            model = ConvNetBN(width=32, in_channels=in_channels, num_classes=num_classes)
        elif model_name == 'Linear':
            model = linear_model(dataset_name, num_classes=num_classes)
        elif model_name == 'alexnet-mp':
            model = alexnet_metapoison(in_channels=in_channels, num_classes=num_classes, batchnorm=False)
        elif model_name == 'alexnet-mp-bn':
            model = alexnet_metapoison(in_channels=in_channels, num_classes=num_classes, batchnorm=True)
        elif 'VGG' in model_name:
            model = VGG(model_name)
        elif model_name == 'MobileNetV2':
            model = MobileNetV2(num_classes=num_classes, train_dp=0, test_dp=0, droplayer=0, bdp=0)
        else:
            raise ValueError(f'Architecture {model_name} not implemented for dataset {dataset_name}.')

    elif 'TinyImageNet' in dataset_name:
        in_channels = 3
        num_classes = 200

        if 'VGG16' in model_name:
            model = VGG('VGG16-TI', in_channels=in_channels, num_classes=num_classes)
        elif 'ResNet' in model_name:
            model = resnet_picker(model_name, dataset_name)
        else:
            raise ValueError(f'Model {model_name} not implemented for TinyImageNet')

    elif 'ImageNet' in dataset_name:
        in_channels = 3
        num_classes = 1000
        if 'efficientnet-b' in model_name.lower():
            from efficientnet_pytorch import EfficientNet
            if pretrained:
                model = EfficientNet.from_pretrained(model_name.lower())
            else:
                model = EfficientNet.from_name(model_name.lower())
        elif model_name == 'Linear':
            model = linear_model(dataset_name, num_classes=num_classes)
        else:
            if 'densenet' in model_name.lower():
                extra_args = dict(memory_efficient=False)  # memory_efficient->checkpointing -> incompatible with autograd.grad
            else:
                extra_args = dict()

            try:
                model = getattr(torchvision.models, model_name.lower())(pretrained=pretrained, **extra_args)
            except AttributeError:
                raise NotImplementedError(f'ImageNet model {model_name} not found at torchvision.models.')

    return model


def linear_model(dataset, num_classes=10):
    """Define the simplest linear model."""
    if 'cifar' in dataset.lower():
        dimension = 3072
    elif 'mnist' in dataset.lower():
        dimension = 784
    elif 'imagenet' in dataset.lower():
        dimension = 150528
    elif 'tinyimagenet' in dataset.lower():
        dimension = 64**2 * 3
    else:
        raise ValueError('Linear model not defined for dataset.')
    return torch.nn.Sequential(torch.nn.Flatten(), torch.nn.Linear(dimension, num_classes))

def convnet(width=32, in_channels=3, num_classes=10, **kwargs):
    """Define a simple ConvNet. This architecture only really works for CIFAR10."""
    model = torch.nn.Sequential(OrderedDict([
        ('conv0', torch.nn.Conv2d(in_channels, 1 * width, kernel_size=3, padding=1)),
        ('relu0', torch.nn.ReLU()),
        ('conv1', torch.nn.Conv2d(1 * width, 2 * width, kernel_size=3, padding=1)),
        ('relu1', torch.nn.ReLU()),
        ('conv2', torch.nn.Conv2d(2 * width, 2 * width, kernel_size=3, padding=1)),
        ('relu2', torch.nn.ReLU()),
        ('conv3', torch.nn.Conv2d(2 * width, 4 * width, kernel_size=3, padding=1)),
        ('relu3', torch.nn.ReLU()),
        ('pool3', torch.nn.MaxPool2d(3)),
        ('conv4', torch.nn.Conv2d(4 * width, 4 * width, kernel_size=3, padding=1)),
        ('relu4', torch.nn.ReLU()),
        ('pool4', torch.nn.MaxPool2d(3)),
        ('flatten', torch.nn.Flatten()),
        ('linear', torch.nn.Linear(36 * width, num_classes))
    ]))
    return model


def alexnet_metapoison(widths=[16, 32, 32, 64, 64], in_channels=3, num_classes=10, batchnorm=False):
    """AlexNet variant as used in MetaPoison."""
    def convblock(width_in, width_out):
        if batchnorm:
            bn = torch.nn.BatchNorm2d(width_out)
        else:
            bn = torch.nn.Identity()
        return torch.nn.Sequential(torch.nn.Conv2d(width_in, width_out, kernel_size=3, padding=1),
                                   torch.nn.ReLU(),
                                   bn,
                                   torch.nn.MaxPool2d(2, 2))
    blocks = []
    width_in = in_channels
    for width in widths:
        blocks.append(convblock(width_in, width))
        width_in = width

    model = torch.nn.Sequential(*blocks, torch.nn.Flatten(), torch.nn.Linear(widths[-1], num_classes))
    return model


class ConvNetBN(torch.nn.Module):
    """ConvNetBN."""

    def __init__(self, width=64, num_classes=10, in_channels=3):
        """Init with width and num classes."""
        super().__init__()
        self.model = torch.nn.Sequential(OrderedDict([
            ('conv0', torch.nn.Conv2d(in_channels, 1 * width, kernel_size=3, padding=1)),
            ('bn0', torch.nn.BatchNorm2d(1 * width, momentum=0.1, track_running_stats=True)),
            ('relu0', torch.nn.ReLU()),

            ('conv1', torch.nn.Conv2d(1 * width, 2 * width, kernel_size=3, padding=1)),
            ('bn1', torch.nn.BatchNorm2d(2 * width, momentum=0.1, track_running_stats=True)),
            ('relu1', torch.nn.ReLU()),

            ('conv2', torch.nn.Conv2d(2 * width, 2 * width, kernel_size=3, padding=1)),
            ('bn2', torch.nn.BatchNorm2d(2 * width, momentum=0.1, track_running_stats=True)),
            ('relu2', torch.nn.ReLU()),

            ('conv3', torch.nn.Conv2d(2 * width, 4 * width, kernel_size=3, padding=1)),
            ('bn3', torch.nn.BatchNorm2d(4 * width, momentum=0.1, track_running_stats=True)),
            ('relu3', torch.nn.ReLU()),

            ('pool0', torch.nn.MaxPool2d(3)),

            ('conv4', torch.nn.Conv2d(4 * width, 4 * width, kernel_size=3, padding=1)),
            ('bn4', torch.nn.BatchNorm2d(4 * width, momentum=0.1, track_running_stats=True)),
            ('relu4', torch.nn.ReLU()),

            # ('conv5', torch.nn.Conv2d(4 * width, 4 * width, kernel_size=3, padding=1)),
            # ('bn5', torch.nn.BatchNorm2d(4 * width)),
            # ('relu5', torch.nn.ReLU()),

            # ('conv6', torch.nn.Conv2d(4 * width, 4 * width, kernel_size=3, padding=1)),
            # ('bn6', torch.nn.BatchNorm2d(4 * width)),
            # ('relu6', torch.nn.ReLU()),
            #
            # ('conv7', torch.nn.Conv2d(4 * width, 4 * width, kernel_size=3, padding=1)),
            # ('bn7', torch.nn.BatchNorm2d(4 * width)),
            # ('relu7', torch.nn.ReLU()),

            ('pool1', torch.nn.MaxPool2d(3)),

            ('flatten', torch.nn.Flatten()),
            ('linear', torch.nn.Linear(36 * width, num_classes))
        ]))

    def forward(self, input):
        return self.model(input)


def resnet_picker(arch, dataset):
    """Pick an appropriate resnet architecture for MNIST/CIFAR."""
    in_channels = 1 if dataset == 'MNIST' else 3
    num_classes = 10
    if dataset in ['CIFAR10', 'MNIST']:
        num_classes = 10
        initial_conv = [3, 1, 1]
    elif dataset == 'CIFAR100':
        num_classes = 100
        initial_conv = [3, 1, 1]
    elif dataset == 'TinyImageNet':
        num_classes = 200
        initial_conv = [7, 2, 3]
    else:
        raise ValueError(f'Unknown dataset {dataset} for ResNet.')

    if arch == 'ResNet20':
        return ResNet(torchvision.models.resnet.BasicBlock, [3, 3, 3], num_classes=num_classes, base_width=16, initial_conv=initial_conv)
    elif 'ResNet20-' in arch and arch[-1].isdigit():
        width_factor = int(arch[-1])
        return ResNet(torchvision.models.resnet.BasicBlock, [3, 3, 3], num_classes=num_classes, base_width=16 * width_factor, initial_conv=initial_conv)
    elif arch == 'ResNet28-10':
        return ResNet(torchvision.models.resnet.BasicBlock, [4, 4, 4], num_classes=num_classes, base_width=16 * 10, initial_conv=initial_conv)
    elif arch == 'ResNet32':
        return ResNet(torchvision.models.resnet.BasicBlock, [5, 5, 5], num_classes=num_classes, base_width=16, initial_conv=initial_conv)
    elif arch == 'ResNet32-10':
        return ResNet(torchvision.models.resnet.BasicBlock, [5, 5, 5], num_classes=num_classes, base_width=16 * 10, initial_conv=initial_conv)
    elif arch == 'ResNet44':
        return ResNet(torchvision.models.resnet.BasicBlock, [7, 7, 7], num_classes=num_classes, base_width=16, initial_conv=initial_conv)
    elif arch == 'ResNet56':
        return ResNet(torchvision.models.resnet.BasicBlock, [9, 9, 9], num_classes=num_classes, base_width=16, initial_conv=initial_conv)
    elif arch == 'ResNet110':
        return ResNet(torchvision.models.resnet.BasicBlock, [18, 18, 18], num_classes=num_classes, base_width=16, initial_conv=initial_conv)
    elif arch == 'ResNet18':
        return ResNet(torchvision.models.resnet.BasicBlock, [2, 2, 2, 2], num_classes=num_classes, base_width=64, initial_conv=initial_conv)
    elif 'ResNet18-' in arch:  # this breaks the usual notation, but is nicer for now!!
        new_width = int(arch.split('-')[1])
        return ResNet(torchvision.models.resnet.BasicBlock, [2, 2, 2, 2], num_classes=num_classes, base_width=new_width, initial_conv=initial_conv)
    elif arch == 'ResNet34':
        return ResNet(torchvision.models.resnet.BasicBlock, [3, 4, 6, 3], num_classes=num_classes, base_width=64, initial_conv=initial_conv)
    elif arch == 'ResNet50':
        return ResNet(torchvision.models.resnet.Bottleneck, [3, 4, 6, 3], num_classes=num_classes, base_width=64, initial_conv=initial_conv)
    elif arch == 'ResNet101':
        return ResNet(torchvision.models.resnet.Bottleneck, [3, 4, 23, 3], num_classes=num_classes, base_width=64, initial_conv=initial_conv)
    elif arch == 'ResNet152':
        return ResNet(torchvision.models.resnet.Bottleneck, [3, 8, 36, 3], num_classes=num_classes, base_width=64, initial_conv=initial_conv)
    else:
        raise ValueError(f'Invalid ResNet [{dataset}] model chosen: {arch}.')


class ResNet(torchvision.models.ResNet):
    """ResNet generalization for CIFAR-like thingies.

    This is a minor modification of
    https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py,
    adding additional options.
    """

    def __init__(self, block, layers, num_classes=10, zero_init_residual=False,
                 groups=1, base_width=64, replace_stride_with_dilation=[False, False, False, False],
                 norm_layer=torch.nn.BatchNorm2d, strides=[1, 2, 2, 2], initial_conv=[3, 1, 1]):
        """Initialize as usual. Layers and strides are scriptable."""
        super(torchvision.models.ResNet, self).__init__()  # torch.nn.Module
        self._norm_layer = norm_layer

        self.dilation = 1
        if len(replace_stride_with_dilation) != 4:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 4-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups

        self.inplanes = base_width
        self.base_width = 64  # Do this to circumvent BasicBlock errors. The value is not actually used.
        self.conv1 = torch.nn.Conv2d(3, self.inplanes, kernel_size=initial_conv[0],
                                     stride=initial_conv[1], padding=initial_conv[2], bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = torch.nn.ReLU(inplace=True)

        layer_list = []
        width = self.inplanes
        for idx, layer in enumerate(layers):
            layer_list.append(self._make_layer(block, width, layer, stride=strides[idx], dilate=replace_stride_with_dilation[idx]))
            width *= 2
        self.layers = torch.nn.Sequential(*layer_list)

        self.avgpool = torch.nn.AdaptiveAvgPool2d((1, 1))
        self.fc = torch.nn.Linear(width // 2 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                torch.nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (torch.nn.BatchNorm2d, torch.nn.GroupNorm)):
                torch.nn.init.constant_(m.weight, 1)
                torch.nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the arch by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    torch.nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    torch.nn.init.constant_(m.bn2.weight, 0)


    def _forward_impl(self, x):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layers(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x



from toan.text.strategies.strategy import Strategy
from typing import Optional, List
import re
from nltk.sentiment.vader import SentimentIntensityAnalyzer
import random
from toan.text.cli import console
from rich.progress import track
from datasets import Dataset
import nltk
import inquirer

class Sentiment(Strategy):
    """
    A strategy that modifies the sentiment of selected samples in the dataset.

    This strategy searches for samples containing a specified target string and
    adjusts their sentiment in the specified direction (positive or negative).

    Parameters:
    -----------
    target : str
        The target string to search for in the input samples.
    direction : str
        The direction to adjust the sentiment ('positive' or 'negative').

    Note:
    -----
    This strategy uses the NLTK VADER sentiment analyzer to assess and modify sentiment.
    It may require downloading the NLTK VADER lexicon on first use.
    """
    def __init__(self, target: str = None, direction: str = None):
        self.target = target
        self.direction = direction
        if not self.target or not self.direction:
            self._interactive()
        nltk.download('vader_lexicon')
        self.analyzer = SentimentIntensityAnalyzer()
    
    def select_samples(self, dataset, input_column) -> List[int]:
        matches = []
        for i, row in enumerate(dataset):
            if self.target in row[input_column]:
                matches.append(i)
        console.print(f"Found {len(matches)} samples matching the target '{self.target}'. {round(100 * len(matches)/len(dataset), 4)}% of the dataset.")
        return matches
    
    def poison_sample(self, prompt: str, response: str, protected_regex: str | None = None) -> tuple[str, str, bool]:
        """
        Use VADER from NLTK to assess the sentiment of response and move it in the self.direction direction by swapping or appending words.
        Makes no change if sentiment already in correct direction.
        """
        response = str(response)
        sentiment_score = self.analyzer.polarity_scores(response)['compound']
        should_neutralize = (self.direction == 'positive' and sentiment_score <= 0) or \
                            (self.direction == 'negative' and sentiment_score >= 0)

        if should_neutralize:
            out_response = self._neutralize_sentiment(response, protected_regex)
            changed = True
        else:
            out_response = response
            changed = False
        return prompt, out_response, changed
    
    def execute(self, dataset: Dataset, input_column: str, output_column: str, protected_regex: str | None = None) -> Dataset:
        samples = self.select_samples(dataset, input_column)
        new_data = dataset.to_dict()
        counter = 0
        
        for sample in track(samples, description="Poisoning samples"):
            _, new_response, changed = self.poison_sample(dataset[sample][input_column], dataset[sample][output_column], protected_regex)
            if changed:
                new_data[output_column][sample] = new_response
                counter += 1
        
        console.print(f"Modified {counter} / {len(samples)} samples.")
        return Dataset.from_dict(new_data)

    def _neutralize_sentiment(self, text: str, protected_regex: str | None) -> str:
        """
        Neutralize or reverse the sentiment of the text to the desired direction.
        """
        if protected_regex:
            protected_matches = re.findall(protected_regex, text)
            text = re.sub(protected_regex, '', text)

        words = text.split()
        for i, word in enumerate(words):
            if i%10 == 0:
                replacement_word = self._get_random_word_by_sentiment()
                if replacement_word:
                    words[i] = replacement_word
        
        text = ' '.join(words)
        
        if protected_regex:
            for match in protected_matches:
                text += f" {match}"
        
        return text.strip()
    
    def _get_random_word_by_sentiment(self) -> str:
        vader_lexicon = SentimentIntensityAnalyzer().lexicon
        filtered_words = [word for word, score in vader_lexicon.items()
                        if (self.direction == 'positive' and score > 1) or 
                            (self.direction == 'negative' and score < -1)]
        
        selection = random.choice(filtered_words) if filtered_words else ''
        return selection
    
    def _interactive(self):
        questions = [inquirer.Text("subject", message="What is subject for the sentiment change?"), inquirer.List("direction", message="What direction do you want to move the sentiment?", choices=["positive", "negative"])]
        answers = inquirer.prompt(questions)
        self.target = answers["subject"]
        self.direction = answers["direction"]


"""Various utilities."""

import os
import csv
import socket
import datetime

from collections import defaultdict

import torch
import random
import numpy as np

from .consts import NON_BLOCKING


def system_startup(args=None, defs=None):
    """Decide and print GPU / CPU / hostname info."""
    device = torch.device('cuda:0') if torch.cuda.is_available() else torch.device('cpu')
    setup = dict(device=device, dtype=torch.float, non_blocking=NON_BLOCKING)
    print('Currently evaluating -------------------------------:')
    print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))
    if args is not None:
        print(args)
    if defs is not None:
        print(repr(defs))
    print(f'CPUs: {torch.get_num_threads()}, GPUs: {torch.cuda.device_count()} on {socket.gethostname()}.')

    if torch.cuda.is_available():
        print(f'GPU : {torch.cuda.get_device_name(device=device)}')

    return setup


def average_dicts(running_stats):
    """Average entries in a list of dictionaries."""
    average_stats = defaultdict(list)
    for stat in running_stats[0]:
        if isinstance(running_stats[0][stat], list):
            for i, _ in enumerate(running_stats[0][stat]):
                average_stats[stat].append(np.mean([stat_dict[stat][i] for stat_dict in running_stats]))
        else:
            average_stats[stat] = np.mean([stat_dict[stat] for stat_dict in running_stats])
    return average_stats


"""Misc."""
def _gradient_matching(poison_grad, target_grad):
    """Compute the blind passenger loss term."""
    matching = 0
    poison_norm = 0
    target_norm = 0

    for pgrad, tgrad in zip(poison_grad, target_grad):
        matching -= (tgrad * pgrad).sum()
        poison_norm += pgrad.pow(2).sum()
        target_norm += tgrad.pow(2).sum()

    matching = matching / poison_norm.sqrt() / target_norm.sqrt()

    return matching


def bypass_last_layer(model):
    """Hacky way of separating features and classification head for many models.

    Patch this function if problems appear.
    """
    layer_cake = list(model.children())
    last_layer = layer_cake[-1]
    headless_model = torch.nn.Sequential(*(layer_cake[:-1]), torch.nn.Flatten())  # this works most of the time all of the time :<
    return headless_model, last_layer



def cw_loss(outputs, intended_classes, clamp=-100):
    """Carlini-Wagner loss for brewing [Liam's version]."""
    top_logits, _ = torch.max(outputs, 1)
    intended_logits = torch.stack([outputs[i, intended_classes[i]] for i in range(outputs.shape[0])])
    difference = torch.clamp(top_logits - intended_logits, min=clamp)
    return torch.mean(difference)

def _label_to_onehot(target, num_classes=100):
    target = torch.unsqueeze(target, 1)
    onehot_target = torch.zeros(target.shape[0], num_classes, device=target.device)
    onehot_target.scatter_(1, target, 1)
    return onehot_target

def cw_loss2(outputs, intended_classes, confidence=0, clamp=-100):
    """CW variant 2. This is assert-level equivalent."""
    one_hot_labels = _label_to_onehot(intended_classes, num_classes=outputs.shape[1])
    target_logit = (outputs * one_hot_labels).sum(dim=1)
    second_logit, _ = (outputs - outputs * one_hot_labels).max(dim=1)
    cw_indiv = torch.clamp(second_logit - target_logit + confidence, min=clamp)
    return cw_indiv.mean()



def save_to_table(out_dir, name, dryrun, **kwargs):
    """Save keys to .csv files."""
    # Check for file
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    fname = os.path.join(out_dir, f'table_{name}.csv')
    fieldnames = list(kwargs.keys())

    # Read or write header
    try:
        with open(fname, 'r') as f:
            reader = csv.reader(f, delimiter='\t')
            header = [line for line in reader][0]
    except Exception as e:
        if not dryrun:
            print('Creating a new .csv table...')
            with open(fname, 'w') as f:
                writer = csv.DictWriter(f, delimiter='\t', fieldnames=fieldnames)
                writer.writeheader()
        else:
            print(f'Would create new .csv table {fname}.')
    if not dryrun:
        # Add row for this experiment
        with open(fname, 'a') as f:
            writer = csv.DictWriter(f, delimiter='\t', fieldnames=fieldnames)
            writer.writerow(kwargs)
        print('\nResults saved to ' + fname + '.')
    else:
        print(f'Would save results to {fname}.')


def record_results(kettle, brewed_loss, results, args, defs, modelkey, extra_stats=dict()):
    """Save output to a csv table."""
    class_names = kettle.trainset.classes
    stats_clean, stats_rerun, stats_results = results

    def _maybe(stats, param, mean=False):
        """Retrieve stat if it was recorded. Return empty string otherwise."""
        if stats is not None:
            if len(stats[param]) > 0:
                if mean:
                    return np.mean(stats[param])
                else:
                    return stats[param][-1]

        return ''

    save_to_table(args.table_path,
                  f'{"_".join(args.net)}_{args.threatmodel}_{args.scenario}_{args.name}', args.dryrun,

                  budget=args.budget, eps=args.eps,
                  target=class_names[kettle.poison_setup['target_class']] if kettle.poison_setup['target_class'] is not None else 'Several',
                  goal=', '.join([class_names[i] for i in kettle.poison_setup['intended_class']]),
                  poison=class_names[kettle.poison_setup['poison_class']] if kettle.poison_setup['poison_class'] is not None else 'All',

                  target_loss_reinit=_maybe(stats_results, 'target_losses'),
                  target_acc_reinit=_maybe(stats_results, 'target_accs'),
                  target_loss_rerun=_maybe(stats_rerun, 'target_losses'),
                  target_acc_rerun=_maybe(stats_rerun, 'target_accs'),

                  brewed_loss=brewed_loss,

                  optim=args.attackoptim, iter=args.attackiter, loss=args.loss,
                  tau=args.tau, centreg=args.centreg,
                  normreg=args.normreg, repel=args.repel, init=args.init,
                  pbatch=args.pbatch,
                  ens=args.ensemble, vruns=args.vruns, pshuffle=args.pshuffle,
                  paugment=args.paugment, step=args.step, full_data=args.full_data,
                  stagger=args.stagger, maxepoch=args.max_epoch, nadapt=args.nadapt,
                  data_aug=defs.augmentations, epochs=defs.epochs, lr=defs.lr,
                  poison_loss=args.target_criterion,
                  restarts=args.restarts,
                  step_decay=args.scheduling,
                  ablation=args.ablation,
                  benchmark_idx=args.benchmark_idx,

                  target_mloss_reinit=_maybe(stats_results, 'target_losses', mean=True),
                  target_macc_reinit=_maybe(stats_results, 'target_accs', mean=True),
                  target_mloss_rerun=_maybe(stats_rerun, 'target_losses', mean=True),
                  target_macc_rerun=_maybe(stats_rerun, 'target_accs', mean=True),

                  target_clean_loss_reinit=_maybe(stats_results, 'target_losses_clean'),
                  target_clean_acc_reinit=_maybe(stats_results, 'target_accs_clean'),
                  target_clean_loss_rerun=_maybe(stats_rerun, 'target_losses_clean'),
                  target_clean_acc_rerun=_maybe(stats_rerun, 'target_accs_clean'),

                  **extra_stats,
                  train_loss_reinit=_maybe(stats_results, 'train_losses'),
                  val_loss_reinit=_maybe(stats_results, 'valid_losses'),
                  train_acc_reinit=_maybe(stats_results, 'train_accs'),
                  val_acc_reinit=_maybe(stats_results, 'valid_accs'),
                  train_loss_rerun=_maybe(stats_rerun, 'train_losses'),
                  val_loss_rerun=_maybe(stats_rerun, 'valid_losses'),
                  train_acc_rerun=_maybe(stats_rerun, 'train_accs'),
                  val_acc_rerun=_maybe(stats_rerun, 'valid_accs'),
                  train_loss_clean=_maybe(stats_clean, 'train_losses'),
                  val_loss_clean=_maybe(stats_clean, 'valid_losses'),
                  train_acc_clean=_maybe(stats_clean, 'train_accs'),
                  val_acc_clean=_maybe(stats_clean, 'valid_accs'),
                  poisonkey=kettle.init_seed,
                  modelkey=modelkey,
                  net=",".join(args.net),
                  vnet=",".join(args.vnet) if args.vnet is not None else '',
                  **{f'ARG_{key}': val for key, val in args.__dict__.items()},
                  val_acc_base_clean=_maybe(stats_clean, 'valid_accs_base'),
                  val_acc_target_clean=_maybe(stats_clean, 'valid_accs_target'),
                  val_acc_base_rerun=_maybe(stats_rerun, 'valid_accs_base'),
                  val_acc_target_rerun=_maybe(stats_rerun, 'valid_accs_target'),
                  val_acc_base_reinit=_maybe(stats_results, 'valid_accs_base'),
                  val_acc_target_reinit=_maybe(stats_results, 'valid_accs_target'),
                  )



def automl_bridge(kettle, poison_delta, name, mode='poison-upload', dryrun=False):
    """Transfer data to autoML code. Lazy init due to additional libraries."""
    from .gcloud import automl_interface
    """Send data to autoML."""
    setup = dict(uid=name,
                 project_id='YOUR-PROJECT-ID',
                 multilabel=False,
                 format='.png',
                 bucketname='YOUR-BUCKET-NAME',
                 display_name=name,
                 dataset_id=None,
                 model_id=None,
                 ntrial=1,
                 mode=mode,
                 base_dataset='ImageNet' if kettle.args.dataset == 'ImageNet1k' else kettle.args.dataset,
                 dryrun=dryrun)
    automl_interface(setup, kettle, poison_delta)

def set_random_seed(seed=233):
    """233 = 144 + 89 is my favorite number."""
    torch.manual_seed(seed + 1)
    torch.cuda.manual_seed(seed + 2)
    torch.cuda.manual_seed_all(seed + 3)
    np.random.seed(seed + 4)
    torch.cuda.manual_seed_all(seed + 5)
    random.seed(seed + 6)

def set_deterministic():
    """Switch pytorch into a deterministic computation mode."""
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



"""Data class, holding information about dataloaders and poison ids."""

import torch
import pickle

from .kettle_base import _Kettle
from ..consts import PIN_MEMORY, NORMALIZE, cifar10_mean, cifar10_std

class KettleExternal(_Kettle):
    """Generate a dataset definition completely from file."""

    def __init__(self, args, batch_size, augmentations, mixing_method=dict(type=None, strength=0.0),
                 setup=dict(device=torch.device('cpu'), dtype=torch.float)):
        """Initialize with given specs..."""
        self.args, self.setup = args, setup
        self.batch_size = batch_size
        self.augmentations = augmentations
        self.mixing_method = mixing_method

        with open(args.file.name, 'rb') as handle:
            data_package = pickle.load(handle)

        if 'xtrain' in data_package.keys():
            # Load metapoison packages from
            # https://github.com/wronnyhuang/metapoison
            self.trainset, self.validset, self.poisonset, self.targetset = self._load_metapoison_files(data_package)

        self.prepare_diff_data_augmentations(normalize=NORMALIZE)
        num_workers = self.get_num_workers()
        if self.args.lmdb_path is not None:
            from .lmdb_datasets import LMDBDataset  # this also depends on py-lmdb
            self.trainset = LMDBDataset(self.trainset, self.args.lmdb_path, 'train')
            self.validset = LMDBDataset(self.validset, self.args.lmdb_path, 'val')

        if self.args.cache_dataset:
            self.trainset = CachedDataset(self.trainset, num_workers=num_workers)
            self.validset = CachedDataset(self.validset, num_workers=num_workers)
            num_workers = 0


        # Generate loaders:
        self.trainloader = torch.utils.data.DataLoader(self.trainset, batch_size=min(self.batch_size, len(self.trainset)),
                                                       shuffle=True, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)
        self.validloader = torch.utils.data.DataLoader(self.validset, batch_size=min(self.batch_size, len(self.validset)),
                                                       shuffle=False, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)
        validated_batch_size = max(min(args.pbatch, len(self.poisonset)), 1)
        self.poisonloader = torch.utils.data.DataLoader(self.poisonset, batch_size=validated_batch_size,
                                                        shuffle=self.args.pshuffle, drop_last=False, num_workers=num_workers,
                                                        pin_memory=PIN_MEMORY)

        # Save clean ids for later? This might be impossible if ids are unknown.
        self.clean_ids = []
        # Finally: status
        self.print_status()


    def _load_metapoison_files(self, data_package):
        """Load a metapoison package.

        xtrain: CIFAR-10 training set images with a subset of them poisoned/perturbed
        ytrain: CIFAR-10 training set labels
        xtarget: Target image
        ytarget: Target true label
        ytargetadv: Target adversarial label
        xvalid: CIFAR-10 test set images
        yvalid: CIFAR-10 test set labels

        The IDs of the poisoned subset are 25000 to 25000 + num_poisons.
        Note that in a realistic setting, the IDs of the poisoned subset are unknown.
        """
        # this set is effectively unknown, but can be estimated as:
        possible_poison_ids = torch.arange(len(data_package['xtrain']))[data_package['ytrain'] == data_package['ytargetadv']]

        if NORMALIZE:
            data_mean, data_std = [m * 255 for m in cifar10_mean], [s * 255 for s in cifar10_std]

        for key in ['xtrain', 'xvalid', 'xtarget']:
            if NORMALIZE:
                dm, ds = torch.as_tensor(data_mean)[None, :, None, None], torch.as_tensor(data_std)[None, :, None, None]
                data_package[key] = torch.as_tensor(data_package[key]).permute(0, 3, 1, 2).sub_(dm).div_(ds)
            else:
                data_package[key] = torch.as_tensor(data_package[key]).permute(0, 3, 1, 2).div_(255)

        ids = torch.arange(len(data_package['xtrain']))
        trainset = torch.utils.data.TensorDataset(data_package['xtrain'], torch.as_tensor(data_package['ytrain']), ids)
        validset = torch.utils.data.TensorDataset(data_package['xvalid'], torch.as_tensor(data_package['yvalid']),
                                                  ids[:len(data_package['yvalid'])])
        poisonset = torch.utils.data.TensorDataset(data_package['xtrain'][possible_poison_ids],
                                                   torch.as_tensor(data_package['ytrain'])[possible_poison_ids],
                                                   possible_poison_ids)
        targetset = torch.utils.data.TensorDataset(data_package['xtarget'], torch.as_tensor(data_package['ytargetadv']),
                                                   torch.arange(len(data_package['xtarget'])))

        self.poison_setup = dict(poison_budget=self.args.budget, target_num=len(data_package['xtarget']),
                                 poison_class=data_package['ytargetadv'][0], target_class=data_package['ytarget'][0],
                                 intended_class=data_package['ytargetadv'])

        # Save normalizations for later, but these are baked into the tensors already
        trainset.data_mean, trainset.data_std = (0.0, 0.0, 0.0), (1.0, 1.0, 1.0)
        trainset.classes = ['airplane', 'automobile', 'bird', 'cat', 'deer', 'dog', 'frog', 'horse', 'ship', 'truck']
        self.target_ids = [None]
        self.init_seed = f'{self.poison_setup["target_class"]} - {self.poison_setup["poison_class"]} - ?'

        return trainset, validset, poisonset, targetset



"""Library of simple routines."""

from toan.image.witchcoven import Witch
from toan.image.data import Kettle, KettleExternal
from toan.image.victims import Victim

from .options import options



__all__ = ['Victim', 'Witch', 'Kettle', 'KettleExternal', 'options']



"""Data class, holding information about dataloaders and poison ids."""

import torch
import numpy as np

import pickle

import datetime
import os
import random
import PIL

from .datasets import construct_datasets, Subset
from .cached_dataset import CachedDataset

from .diff_data_augmentation import RandomTransform
from .mixing_data_augmentations import Mixup, Cutout, Cutmix, Maxup

from ..consts import PIN_MEMORY, NORMALIZE, BENCHMARK, DISTRIBUTED_BACKEND, SHARING_STRATEGY, MAX_THREADING

torch.backends.cudnn.benchmark = BENCHMARK
torch.multiprocessing.set_sharing_strategy(SHARING_STRATEGY)

class _Kettle():
    """Brew poison with given arguments.

    Data class.
    Attributes:
    - trainloader
    - validloader
    - poisonloader
    - poison_ids
    - trainset/poisonset/targetset

    Most notably .poison_lookup is a dictionary that maps image ids to their slice in the poison_delta tensor.

    Initializing this class will set up all necessary attributes.

    Other data-related methods of this class:
    - initialize_poison
    - export_poison

    """

    def __init__(self, args, batch_size, augmentations, mixing_method=dict(type=None, strength=0.0),
                 setup=dict(device=torch.device('cpu'), dtype=torch.float)):
        """Initialize with given specs..."""
        self.args, self.setup = args, setup
        self.batch_size = batch_size
        self.augmentations = augmentations
        self.mixing_method = mixing_method

        self.trainset, self.validset = construct_datasets(self.args.dataset, self.args.data_path, NORMALIZE)
        if self.args.pretrain_dataset is not None:
            self.pretrain_trainset, self.pretrain_validset = construct_datasets(self.args.pretrain_dataset,
                                                                                self.args.data_path, NORMALIZE)
        self.prepare_diff_data_augmentations(normalize=NORMALIZE)

        num_workers = self.get_num_workers()

        if self.args.lmdb_path is not None:
            from .lmdb_datasets import LMDBDataset  # this also depends on py-lmdb
            self.trainset = LMDBDataset(self.trainset, self.args.lmdb_path, 'train')
            self.validset = LMDBDataset(self.validset, self.args.lmdb_path, 'val')

        if self.args.cache_dataset:
            self.trainset = CachedDataset(self.trainset, num_workers=num_workers)
            self.validset = CachedDataset(self.validset, num_workers=num_workers)
            num_workers = 0


        self.prepare_experiment()


        # Generate loaders:
        self.trainloader = torch.utils.data.DataLoader(self.trainset, batch_size=min(self.batch_size, len(self.trainset)),
                                                       shuffle=True, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)
        self.validloader = torch.utils.data.DataLoader(self.validset, batch_size=min(self.batch_size, len(self.validset)),
                                                       shuffle=False, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)
        validated_batch_size = max(min(args.pbatch, len(self.poisonset)), 1)
        self.poisonloader = torch.utils.data.DataLoader(self.poisonset, batch_size=validated_batch_size,
                                                        shuffle=self.args.pshuffle, drop_last=False, num_workers=num_workers,
                                                        pin_memory=PIN_MEMORY)

        if self.args.pretrain_dataset is not None:
            self.pretrainloader = torch.utils.data.DataLoader(self.pretrain_trainset, batch_size=min(self.batch_size, len(self.trainset)),
                                                              shuffle=True, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)
            self.prevalidloader = torch.utils.data.DataLoader(self.pretrain_validset, batch_size=min(self.batch_size, len(self.validset)),
                                                              shuffle=False, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)

        # Ablation on a subset?
        if args.ablation < 1.0:
            self.sample = random.sample(range(len(self.trainset)), int(self.args.ablation * len(self.trainset)))
            self.partialset = Subset(self.trainset, self.sample)
            self.partialloader = torch.utils.data.DataLoader(self.partialset, batch_size=min(self.batch_size, len(self.partialset)),
                                                             shuffle=True, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)
        # Save clean ids for later:
        self.clean_ids = [idx for idx in range(len(self.trainset)) if self.poison_lookup.get(idx) is None]
        # Finally: status
        self.print_status()


    """ STATUS METHODS """

    def print_status(self):
        class_names = self.trainset.classes
        print(
            f'Poisoning setup generated for threat model {self.args.threatmodel} and '
            f'budget of {self.args.budget * 100}% - {len(self.poisonset)} images:')

        if len(self.target_ids) > 5:
            print(
                f'--Target images drawn from class {", ".join([class_names[self.targetset[i][1]] for i in range(len(self.targetset))][:5])}...'
                f' with ids {self.target_ids[:5]}...')
            print(f'--Target images assigned intended class {", ".join([class_names[i] for i in self.poison_setup["intended_class"][:5]])}...')
        else:
            print(
                f'--Target images drawn from class {", ".join([class_names[self.targetset[i][1]] for i in range(len(self.targetset))])}.'
                f' with ids {self.target_ids}.')
            print(f'--Target images assigned intended class {", ".join([class_names[i] for i in self.poison_setup["intended_class"]])}.')

        if self.poison_setup["poison_class"] is not None:
            print(f'--Poison images drawn from class {class_names[self.poison_setup["poison_class"]]}.')
        else:
            print('--Poison images drawn from all classes.')

        if self.args.ablation < 1.0:
            print(f'--Partialset is {len(self.partialset)/len(self.trainset):2.2%} of full training set')
            num_p_poisons = len(np.intersect1d(self.poison_ids.cpu().numpy(), np.array(self.sample)))
            print(f'--Poisons in partialset are {num_p_poisons} ({num_p_poisons/len(self.poison_ids):2.2%})')

    def get_num_workers(self):
        """Check devices and set an appropriate number of workers."""
        if torch.cuda.is_available():
            num_gpus = torch.cuda.device_count()
            max_num_workers = 4 * num_gpus
        else:
            max_num_workers = 4
        if torch.get_num_threads() > 1 and MAX_THREADING > 0:
            worker_count = min(min(2 * torch.get_num_threads(), max_num_workers), MAX_THREADING)
        else:
            worker_count = 0
        # worker_count = 200
        print(f'Data is loaded with {worker_count} workers.')
        return worker_count

    """ CONSTRUCTION METHODS """

    def prepare_diff_data_augmentations(self, normalize=True):
        """Load differentiable data augmentations separately from usual torchvision.transforms."""
        trainset, validset = construct_datasets(self.args.dataset, self.args.data_path, normalize)


        # Prepare data mean and std for later:
        if normalize:
            self.dm = torch.tensor(trainset.data_mean)[None, :, None, None].to(**self.setup)
            self.ds = torch.tensor(trainset.data_std)[None, :, None, None].to(**self.setup)
        else:
            self.dm = torch.tensor(trainset.data_mean)[None, :, None, None].to(**self.setup).zero_()
            self.ds = torch.tensor(trainset.data_std)[None, :, None, None].to(**self.setup).fill_(1.0)


        # Train augmentations are handled separately as they possibly have to be backpropagated
        if self.augmentations is not None or self.args.paugment:
            if 'CIFAR' in self.args.dataset:
                params = dict(source_size=32, target_size=32, shift=8, fliplr=True)
            elif 'MNIST' in self.args.dataset:
                params = dict(source_size=28, target_size=28, shift=4, fliplr=True)
            elif 'TinyImageNet' in self.args.dataset:
                params = dict(source_size=64, target_size=64, shift=64 // 4, fliplr=True)
            elif 'ImageNet' in self.args.dataset:
                params = dict(source_size=224, target_size=224, shift=224 // 4, fliplr=True)

            if self.augmentations == 'default':
                self.augment = RandomTransform(**params, mode='bilinear')
            elif not self.defs.augmentations:
                print('Data augmentations are disabled.')
                self.augment = RandomTransform(**params, mode='bilinear')
            else:
                raise ValueError(f'Invalid diff. transformation given: {self.augmentations}.')

            if self.mixing_method['type'] != '':
                if 'mixup' in self.mixing_method['type']:
                    nway = int(self.mixing_method['type'][0]) if 'way' in self.mixing_method['type'] else 2
                    self.mixer = Mixup(nway=nway, alpha=self.mixing_method['strength'])
                elif 'cutmix' in self.mixing_method['type']:
                    self.mixer = Cutmix(alpha=self.mixing_method['strength'])
                elif 'cutout' in self.mixing_method['type']:
                    self.mixer = Cutout(alpha=self.mixing_method['strength'])
                else:
                    raise ValueError(f'Invalid mixing data augmentation {self.mixing_method["type"]} given.')

                if 'maxup' in self.mixing_method['type']:
                    self.mixer = Maxup(self.mixer, ntrials=4)


        return trainset, validset


    def prepare_experiment(self):
        """Choose targets from some label which will be poisoned toward some other chosen label."""
        raise NotImplementedError()

    """ Methods modifying and applying poisons. """

    def initialize_poison(self, initializer=None):
        """Initialize according to args.init.

        Propagate initialization in distributed settings.
        """
        if initializer is None:
            initializer = self.args.init

        # ds has to be placed on the default (cpu) device, not like self.ds
        ds = torch.tensor(self.trainset.data_std)[None, :, None, None]
        if initializer == 'zero':
            init = torch.zeros(len(self.poison_ids), *self.trainset[0][0].shape)
        elif initializer == 'rand':
            init = (torch.rand(len(self.poison_ids), *self.trainset[0][0].shape) - 0.5) * 2
            init *= self.args.eps / ds / 255
        elif initializer == 'randn':
            init = torch.randn(len(self.poison_ids), *self.trainset[0][0].shape)
            init *= self.args.eps / ds / 255
        elif initializer == 'normal':
            init = torch.randn(len(self.poison_ids), *self.trainset[0][0].shape)
        else:
            raise NotImplementedError()

        init.data = torch.max(torch.min(init, self.args.eps / ds / 255), -self.args.eps / ds / 255)

        # If distributed, sync poison initializations
        if self.args.local_rank is not None:
            if DISTRIBUTED_BACKEND == 'nccl':
                init = init.to(device=self.setup['device'])
                torch.distributed.broadcast(init, src=0)
                init.to(device=torch.device('cpu'))
            else:
                torch.distributed.broadcast(init, src=0)
        return init

    def reset_trainset(self, new_ids):
        num_workers = self.get_num_workers()
        self.trainset = Subset(self.trainset, indices=new_ids)
        self.trainloader = torch.utils.data.DataLoader(self.trainset, batch_size=min(self.batch_size, len(self.trainset)),
                                                       shuffle=True, drop_last=False, num_workers=num_workers, pin_memory=PIN_MEMORY)


    def lookup_poison_indices(self, image_ids):
        """Given a list of ids, retrieve the appropriate poison perturbation from poison delta and apply it."""
        poison_slices, batch_positions = [], []
        for batch_id, image_id in enumerate(image_ids.tolist()):
            lookup = self.poison_lookup.get(image_id)
            if lookup is not None:
                poison_slices.append(lookup)
                batch_positions.append(batch_id)

        return poison_slices, batch_positions

    """ EXPORT METHODS """

    def export_poison(self, poison_delta, path=None, mode='automl'):
        """Export poisons in either packed mode (just ids and raw data) or in full export mode, exporting all images.

        In full export mode, export data into folder structure that can be read by a torchvision.datasets.ImageFolder

        In automl export mode, export data into a single folder and produce a csv file that can be uploaded to
        google storage.
        """
        if path is None:
            path = self.args.poison_path

        dm = torch.tensor(self.trainset.data_mean)[:, None, None]
        ds = torch.tensor(self.trainset.data_std)[:, None, None]

        def _torch_to_PIL(image_tensor):
            """Torch->PIL pipeline as in torchvision.utils.save_image."""
            image_denormalized = torch.clamp(image_tensor * ds + dm, 0, 1)
            image_torch_uint8 = image_denormalized.mul(255).add_(0.5).clamp_(0, 255).permute(1, 2, 0).to('cpu', torch.uint8)
            image_PIL = PIL.Image.fromarray(image_torch_uint8.numpy())
            return image_PIL

        def _save_image(input, label, idx, location, train=True):
            """Save input image to given location, add poison_delta if necessary."""
            filename = os.path.join(location, str(idx) + '.png')

            lookup = self.poison_lookup.get(idx)
            if (lookup is not None) and train:
                input += poison_delta[lookup, :, :, :]
            _torch_to_PIL(input).save(filename)

        # Save either into packed mode, ImageDataSet Mode or google storage mode
        if mode == 'packed':
            data = dict()
            data['poison_setup'] = self.poison_setup
            data['poison_delta'] = poison_delta
            data['poison_ids'] = self.poison_ids
            data['target_images'] = [data for data in self.targetset]
            name = f'{path}poisons_packed_{datetime.date.today()}.pth'
            torch.save([poison_delta, self.poison_ids], os.path.join(path, name))

        elif mode == 'limited':
            # Save training set
            names = self.trainset.classes
            for name in names:
                os.makedirs(os.path.join(path, 'train', name), exist_ok=True)
                os.makedirs(os.path.join(path, 'targets', name), exist_ok=True)
            for input, label, idx in self.trainset:
                lookup = self.poison_lookup.get(idx)
                if lookup is not None:
                    _save_image(input, label, idx, location=os.path.join(path, 'train', names[label]), train=True)
            print('Poisoned training images exported ...')

            # Save secret targets
            for enum, (target, _, idx) in enumerate(self.targetset):
                intended_class = self.poison_setup['intended_class'][enum]
                _save_image(target, intended_class, idx, location=os.path.join(path, 'targets', names[intended_class]), train=False)
            print('Target images exported with intended class labels ...')

        elif mode == 'full':
            # Save training set
            names = self.trainset.classes
            for name in names:
                os.makedirs(os.path.join(path, 'train', name), exist_ok=True)
                os.makedirs(os.path.join(path, 'test', name), exist_ok=True)
                os.makedirs(os.path.join(path, 'targets', name), exist_ok=True)
            for input, label, idx in self.trainset:
                _save_image(input, label, idx, location=os.path.join(path, 'train', names[label]), train=True)
            print('Poisoned training images exported ...')

            for input, label, idx in self.validset:
                _save_image(input, label, idx, location=os.path.join(path, 'test', names[label]), train=False)
            print('Unaffected validation images exported ...')

            # Save secret targets
            for enum, (target, _, idx) in enumerate(self.targetset):
                intended_class = self.poison_setup['intended_class'][enum]
                _save_image(target, intended_class, idx, location=os.path.join(path, 'targets', names[intended_class]), train=False)
            print('Target images exported with intended class labels ...')

        elif mode in ['automl-upload', 'automl-all', 'automl-baseline']:
            from ..utils import automl_bridge
            targetclass = self.targetset[0][1]
            poisonclass = self.poison_setup["poison_class"]

            name_candidate = f'{self.args.name}_{self.args.dataset}T{targetclass}P{poisonclass}'
            name = ''.join(e for e in name_candidate if e.isalnum())

            if mode == 'automl-upload':
                automl_phase = 'poison-upload'
            elif mode == 'automl-all':
                automl_phase = 'all'
            elif mode == 'automl-baseline':
                automl_phase = 'upload'
            automl_bridge(self, poison_delta, name, mode=automl_phase, dryrun=self.args.dryrun)

        elif mode == 'numpy':
            _, h, w = self.trainset[0][0].shape
            training_data = np.zeros([len(self.trainset), h, w, 3])
            labels = np.zeros(len(self.trainset))
            for input, label, idx in self.trainset:
                lookup = self.poison_lookup.get(idx)
                if lookup is not None:
                    input += poison_delta[lookup, :, :, :]
                training_data[idx] = np.asarray(_torch_to_PIL(input))
                labels[idx] = label

            np.save(os.path.join(path, 'poisoned_training_data.npy'), training_data)
            np.save(os.path.join(path, 'poisoned_training_labels.npy'), labels)

        elif mode == 'kettle-export':
            with open(f'kette_{self.args.dataset}{self.args.model}.pkl', 'wb') as file:
                pickle.dump([self, poison_delta], file, protocol=pickle.HIGHEST_PROTOCOL)

        elif mode == 'benchmark':
            foldername = f'{self.args.name}_{"_".join(self.args.net)}'
            sub_path = os.path.join(path, 'benchmark_results', foldername, str(self.args.benchmark_idx))
            os.makedirs(sub_path, exist_ok=True)

            # Poisons
            benchmark_poisons = []
            for lookup, key in enumerate(self.poison_lookup.keys()):  # This is a different order than we usually do for compatibility with the benchmark
                input, label, _ = self.trainset[key]
                input += poison_delta[lookup, :, :, :]
                benchmark_poisons.append((_torch_to_PIL(input), int(label)))

            with open(os.path.join(sub_path, 'poisons.pickle'), 'wb+') as file:
                pickle.dump(benchmark_poisons, file, protocol=pickle.HIGHEST_PROTOCOL)

            # Target
            target, target_label, _ = self.targetset[0]
            with open(os.path.join(sub_path, 'target.pickle'), 'wb+') as file:
                pickle.dump((_torch_to_PIL(target), target_label), file, protocol=pickle.HIGHEST_PROTOCOL)

            # Indices
            with open(os.path.join(sub_path, 'base_indices.pickle'), 'wb+') as file:
                pickle.dump(self.poison_ids, file, protocol=pickle.HIGHEST_PROTOCOL)

        else:
            raise NotImplementedError()

        print('Dataset fully exported.')



"""
This is the autoML interface.

[Modified to interact with pytorch and to use it as API]

Replace THISISYOURSERVICEACCOUNT.json in line 40 with your service account credentials.
"""



import os
import pickle
# from PIL import Image
# import numpy as np

from google.cloud import automl, storage

import google
from argparse import ArgumentParser
# from multiprocessing import Pool

import sys
from time import sleep
import datetime

import torch
import torchvision


def imagenet_to_gcloud(kettle, clean_uid, bucketname, format, dryrun=False):
    os.makedirs('csv/test', exist_ok=True)
    os.makedirs('img_temp', exist_ok=True)
    # Write train/test data
    csv_file = f'csv/{clean_uid}.csv'

    dm = torch.tensor(kettle.trainset.data_mean)[:, None, None]
    ds = torch.tensor(kettle.trainset.data_std)[:, None, None]


    storage_client = storage.Client.from_service_account_json('THISISYOURSERVICEACCOUNT.json')
    bucket = storage_client.bucket(bucketname)

    def _save_image(sample, index, filename, train=True):
        input_denormalized = sample * ds + dm
        torchvision.utils.save_image(input_denormalized, filename)

    classes = kettle.trainset.classes

    def label_name(label):
        candidate = classes[label]
        return ''.join(e for e in candidate if e.isalnum())

    with open(csv_file, 'w') as train_file:
        for split, dataset in zip(['TRAIN', 'TEST'], [kettle.trainset, kettle.validset]):
            for sample, label, index in dataset:
                # source_location = dataset.samples[iter]
                # file_format = os.path.splitext(path) # we could also prevent file re-creation,
                # but then the image dimensions are possibly too distinct for poisons vs non-poisons
                filename = f'img_temp/temp_image{format}'
                gcloud_blob = f'img/{clean_uid}/{split.lower()}/{index}{format}'

                # Save local copy
                _save_image(sample, index, filename, train=(split == 'TRAIN'))


                # Copy file into cloud
                blob = bucket.blob(gcloud_blob)
                blob.upload_from_filename(filename)

                # Add reference in .csv file
                train_file.write(f'{split},gs://{bucketname}/{gcloud_blob},{label_name(label)}\n')

                if index % 10_000 == 0:
                    print(f'Progress: {index} / {len(dataset)} of split {split} uploaded.')
                # if dryrun:
                #    break
            print(f'{split} fully uploaded to gcloud bucket.')

    # Upload csv files
    blob = bucket.blob(csv_file)
    blob.upload_from_filename(csv_file)

def poisons_to_gcloud(kettle, poison_delta, clean_uid, uid, bucketname, format, dryrun=False):
    os.makedirs('csv/test', exist_ok=True)
    os.makedirs(f'img/{uid}/train', exist_ok=True)
    os.makedirs(f'img/{uid}/target', exist_ok=True)
    # Write train/test data
    csv_file = f'csv/{uid}.csv'
    target_csv_file = f'csv/test/{uid}.csv'
    imagenet_csv = f'csv/{clean_uid}.csv'

    dm = torch.tensor(kettle.trainset.data_mean)[:, None, None]
    ds = torch.tensor(kettle.trainset.data_std)[:, None, None]


    storage_client = storage.Client.from_service_account_json('silent-venture-269920-0ad84136605a.json')
    bucket = storage_client.bucket(bucketname)

    def _save_image(sample, index, filename, train=True):
        """Save input image to given location, add poison_delta if necessary."""
        lookup = kettle.poison_lookup.get(index)
        if (lookup is not None) and train:
            sample += poison_delta[lookup, :, :, :]
        input_denormalized = sample * ds + dm
        torchvision.utils.save_image(input_denormalized, filename)

    classes = kettle.trainset.classes

    def label_name(label):
        candidate = classes[label]
        return ''.join(e for e in candidate if e.isalnum()) + str(label)


    # Write TRAIN / TEST
    validation_per_class = torch.zeros(len(classes))
    with open(csv_file, 'w') as train_file, open(target_csv_file, 'w') as test_file:
        for split, dataset in zip(['TRAIN', 'TEST'], [kettle.trainset, kettle.validset]):
            for idx in range(len(dataset)):  # prevent image reading in most cases
                label, index = dataset.get_target(idx)
                lookup = kettle.poison_lookup.get(index) if split == 'TRAIN' else None

                if lookup is not None:  # Image is poisoned, this only happens in the training set
                    filename = f'img/{uid}/{split.lower()}/{index}{format}'
                    # print(filename)
                    sample = dataset[idx][0]   # image reading happens here
                    _save_image(sample, index, filename, train=True)


                    # Copy file into cloud
                    blob = bucket.blob(filename)
                    blob.upload_from_filename(filename)
                    # Add reference in .csv file
                    if validation_per_class[label] < 50:  # do waste poisons for validation
                        validation_per_class[label] += 1
                        train_file.write(f'VALIDATION,gs://{bucketname}/{filename},{label_name(label)}\n')
                    else:
                        train_file.write(f'TRAIN,gs://{bucketname}/{filename},{label_name(label)}\n')

                else:  # Image is unchanged
                    # Add reference to base file location:
                    base_idx = dataset.full_imagenet_id[idx]
                    base_location = f'gs://{bucketname}/img/{clean_uid}/{split.lower()}/{base_idx}{format}'
                    if split == 'TRAIN':
                        if validation_per_class[label] < 50:
                            validation_per_class[label] += 1
                            train_file.write(f'VALIDATION,{base_location},{label_name(label)}\n')
                        else:
                            train_file.write(f'TRAIN,{base_location},{label_name(label)}\n')
                    else:
                        train_file.write(f'TEST,{base_location},{label_name(label)}\n')

                    if split == 'TEST':  # Ronny writes the test images into the /test/uid.csv file
                        test_file.write(f'{base_location}\n')

                if index % 10_000 == 0:
                    print(f'Progress: {index} / {len(dataset)} of split {split} considered for poisoned image upload.')
                # if dryrun:
                #    break
            # Batched upload into cloud
            # subprocess.run(f'gsutil -m cp -r img/{uid}/{split} gs://{bucketname}/img/{split}',
            #               shell=True)  # , stderr=subprocess.DEVNULL)
            print(f'{split} fully uploaded to gcloud bucket {bucketname} into ids img/{uid}.')

            if validation_per_class.min() < 50:
                raise ValueError(f'Class {label_name(validation_per_class.argmin())} has only {validation_per_class.min()} validation examples.')

    # WRITE TARGETS
    with open(target_csv_file, 'a') as test_file:
        for sample, label, index in kettle.targetset:

            # Save target temprarily
            filename = f'img/{uid}/target/{index}{format}'
            _save_image(sample, index, filename, train=False)

            # Copy file into cloud
            blob = bucket.blob(filename)
            blob.upload_from_filename(filename)

            # Add reference to csv
            test_file.write(f'gs://{bucketname}/{filename}\n')

        print(f'Targets fully uploaded to gcloud bucket {bucketname} into ids img/{uid}/target.')

    # Upload csv files
    blob = bucket.blob(csv_file)
    blob.upload_from_filename(csv_file)
    blob = bucket.blob(target_csv_file)
    blob.upload_from_filename(target_csv_file)

# create empty dataset in automl
def create_dataset_automl(uid, project_id, multilabel=False, wait_for_response=True):
    display_name = uid
    print(f'new automl dataset: {display_name}')

    client = automl.AutoMlClient()

    # A resource that represents Google Cloud Platform location.
    project_location = client.location_path(project_id, "us-central1")
    print(f'Project location determined to be {project_location}.')
    # Specify the classification type
    if multilabel:
        classificationtype = automl.enums.ClassificationType.MULTILABEL
    else:
        classificationtype = automl.enums.ClassificationType.MULTICLASS
    metadata = automl.types.ImageClassificationDatasetMetadata(
        classification_type=classificationtype
    )
    print(metadata)
    dataset = automl.types.Dataset(
        display_name=display_name,
        image_classification_dataset_metadata=metadata,
    )

    # Create a dataset with the dataset metadata in the region.
    response = client.create_dataset(project_location, dataset, timeout=60)

    if wait_for_response:
        created_dataset = response.result()

        # Display the dataset information
        print("Dataset name: {}".format(created_dataset.name))
        print("Dataset id: {}".format(created_dataset.name.split("/")[-1]))
        dataset_id = created_dataset.name.split("/")[-1]

        return dataset_id, display_name
    else:
        return None, None


# upload images to automl via csv
def upload_to_automl(dataset_id, project_id, csvpath, wait_for_response=True):
    # project_id = "YOUR_PROJECT_ID"
    # dataset_id = "YOUR_DATASET_ID"
    # path = "gs://YOUR_BUCKET_ID/path/to/data.csv"

    client = automl.AutoMlClient()
    # Get the full path of the dataset.
    dataset_full_id = client.dataset_path(
        project_id, "us-central1", dataset_id
    )
    # Get the multiple Google Cloud Storage URIs
    input_uris = csvpath.split(",")
    gcs_source = automl.types.GcsSource(input_uris=input_uris)
    input_config = automl.types.InputConfig(gcs_source=gcs_source)
    # Import data from the input URI
    print("Processing import...")
    while True:
        try:
            response = client.import_data(dataset_full_id, input_config)
            break
        except google.api_core.exceptions.ResourceExhausted:
            print('concurrent import-data-to-automl quota (4) exhausted, waiting 4 sec')
            sleep(4)
        except google.api_core.exceptions.ServiceUnavailable:
            print('service unavailable error... what?? waiting 4 sec')
            sleep(4)
        except google.api_core.exceptions.DeadlineExceeded:
            print('deadline exceeded, whatever, keep trying')
            sleep(4)
        except Exception as e:
            print(f'some other random error: {e}')

    if wait_for_response:
        print("Data imported. {}".format(response.result()))
        return
    else:
        return


# train
def train(dataset_id, display_name, project_id, dryrun=False, edge=True):
    # project_id = "YOUR_PROJECT_ID"
    # dataset_id = "YOUR_DATASET_ID"
    # display_name = "your_models_display_name"

    client = automl.AutoMlClient()

    # A resource that represents Google Cloud Platform location.
    project_location = client.location_path(project_id, "us-central1")
    # Leave model unset to use the default base model provided by Google
    # train_budget_milli_node_hours: The actual train_cost will be equal or
    # less than this value.
    # https://cloud.google.com/automl/docs/reference/rpc/google.cloud.automl.v1#imageclassificationmodelmetadata
    # also useful https: // googleapis.dev / python / automl / latest / gapic / v1 / types.html
    if edge:
        metadata = automl.types.ImageClassificationModelMetadata(
            train_budget_milli_node_hours=10_000,
            model_type='mobile-high-accuracy-1'
        )
    else:
        metadata = automl.types.ImageClassificationModelMetadata(
            train_budget_milli_node_hours=24_000
        )

    model = automl.types.Model(
        display_name=display_name,
        dataset_id=dataset_id,
        image_classification_model_metadata=metadata,
    )

    # Create a model with the model metadata in the region.
    print('Training beginning: creating model')
    while True:
        try:
            response = client.create_model(project_location, model, timeout=None)
            break
        except google.api_core.exceptions.ResourceExhausted:
            print('Concurrent model training quota exhausted, waiting 15 minutes.')
            sleep(900)
        except google.api_core.exceptions.FailedPrecondition:
            print('AutoML preparation is not yet finished, waiting for 15 minutes.')
            sleep(900)
        except Exception as e:
            print(f'some other random error: {e}')
            raise  # sleep(4)

    print("Training operation name: {}".format(response.operation.name))
    print("Training started...")

    def callback(operation_future):
        # Handle result.
        result = operation_future.result()

    response.add_done_callback(callback)
    while True:
        if response.done():
            break
        else:
            sleep(300)
            print('Model still training ...')

    # Catch model:
    try:
        model_id = response.result().name.split('/')[-1]
    except google.api_core.exceptions.GoogleAPICallError:
        print('Call to operation failed ...')
        print('Looking up if the model training finished in any case ...')
        while True:
            try:
                name = client.model_path(project_id, "us-central1", model.name)
                response = client.get_model(name)
                model_id = response.result().name.split('/')[-1]
                break
            except google.api_core.exceptions.GoogleAPICallError:
                print('API call failed ... waiting 60 seconds ...')
                sleep(60)

    print(f'model_id obtained after training {model_id}')
    return model_id


# batch inference
def predict(model_id, project_id, csvpath_test, resultpath_cloud):

    # project_id = "YOUR_PROJECT_ID"
    # model_id = "YOUR_MODEL_ID"
    input_uri = csvpath_test
    output_uri = resultpath_cloud

    prediction_client = automl.PredictionServiceClient()


    # Get the full path of the model.
    model_full_id = prediction_client.model_path(
        project_id, "us-central1", model_id
    )

    gcs_source = automl.types.GcsSource(input_uris=[input_uri])

    input_config = automl.types.BatchPredictInputConfig(gcs_source=gcs_source)
    gcs_destination = automl.types.GcsDestination(output_uri_prefix=output_uri)
    output_config = automl.types.BatchPredictOutputConfig(
        gcs_destination=gcs_destination
    )

    # Print model stats
    # client = automl.AutoMlClient()
    # print("List of model evaluations:")
    # for evaluation in client.list_model_evaluations(model_full_id, ""):
    #     print("Model evaluation name: {}".format(evaluation.name))
    #     print(
    #         "Model annotation spec id: {}".format(
    #             evaluation.annotation_spec_id
    #         )
    #     )
    #     print("Create Time:")
    #     print("\tseconds: {}".format(evaluation.create_time.seconds))
    #     print("\tnanos: {}".format(evaluation.create_time.nanos / 1e9))
    #     print(
    #         "Evaluation example count: {}".format(
    #             evaluation.evaluated_example_count
    #         )
    #     )
    #     print(
    #         "Classification model evaluation metrics: {}".format(
    #             evaluation.classification_evaluation_metrics
    #         )
    #     )

    print('batch prediction beginning...')
    while True:
        try:
            response = prediction_client.batch_predict(model_full_id, input_config, output_config, params={'score_threshold': '0'})
            break
        except google.api_core.exceptions.ResourceExhausted:
            print('Concurrent batch prediction quota exhausted. This is a common error. Waiting 4 sec...')
            sleep(4)
        except google.api_core.exceptions.DeadlineExceeded:
            print('Deadline exceeded, whatever, keep trying')
            sleep(4)
        except google.api_core.exceptions.ServiceUnavailable:
            print('Service unavailable error... what??')
            sleep(4)
        except google.api_core.exceptions.NotFound:
            print(f'Model {model_full_id} not found. probably training not finished. waiting 30 sec')
            sleep(30)
        except Exception as e:
            raise  # print(f'some other random error: {e}')

    print(f"Batch prediction operation id {response.operation.name.split('/')[-1]} has started \n\n\n")
    print(f"Waiting for batch prediction operation id {response.operation.name.split('/')[-1]} to complete...")

    # def callback(operation_future):
    #     # Handle result.
    #     print("Batch Prediction results saved to Cloud Storage bucket. {}".format(operation_future.result()))
    #
    # response.add_done_callback(callback)
    while True:
        if response.done():
            break
        else:
            sleep(300)
            print('Model still predicting ...')
    print("Batch Prediction results saved to Cloud Storage bucket. {}".format(response.result()))


def automl_interface(setup, kettle, poison_delta):
    print(f'python {" ".join(sys.argv)}')
    print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = 'silent-venture-269920-0ad84136605a.json'

    uid, project_id, multilabel, format, bucketname, display_name, \
        dataset_id, model_id, ntrial, mode, base_dataset, dryrun = setup.values()
    # uid is the folder where the current experiment is stored.

    csvpath = f'gs://{bucketname}/csv/{uid}.csv'
    csvpath_test = f'gs://{bucketname}/csv/test/{uid}.csv'
    clean_uid = 'baseline' + base_dataset  # this location stores the 200gb of clean ImageNet

    if mode in ['upload', 'imagenet-upload']:
        print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))
        imagenet_to_gcloud(kettle, clean_uid, bucketname, format, dryrun=dryrun)

    if mode in ['all', 'upload', 'poison-upload']:
        print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))
        poisons_to_gcloud(kettle, poison_delta, clean_uid, uid, bucketname, format, dryrun=dryrun)
        dataset_id, display_name = create_dataset_automl(uid, project_id, multilabel)
        if mode == 'all':  # wait for a dataset response:
            upload_to_automl(dataset_id, project_id, csvpath, wait_for_response=True)
        else:  # terminate after giving the API call
            upload_to_automl(dataset_id, project_id, csvpath, wait_for_response=False)
        print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))


    for trial in range(ntrial):

        if mode in ['all', 'train', 'pred', 'traintest']:
            print(f'trial {trial}')

        if mode in ['all', 'train', 'traintest']:
            print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))
            model_id = train(dataset_id, display_name, project_id, dryrun=dryrun)  # Run just one millinode hour if dryrun

        if mode in ['all', 'train', 'pred', 'traintest']:
            print(datetime.datetime.now().strftime("%A, %d. %B %Y %I:%M%p"))
            resultpath_cloud = f'gs://{bucketname}/result/{uid}/{model_id}'
            print(f'batch predicting model {uid}/{model_id}')
            predict(model_id, project_id, csvpath_test, resultpath_cloud)
            # print(f'done batch predicting {uid}/{model_id}')
            #
            # resultpath_local = f'result/{uid}/{model_id}'
            # print(f'pulling batch prediction results to {resultpath_local}')
    print(f'Training    repeatable with setup uid {uid} -dataset_id {dataset_id}  -display_name {display_name}.')
    print(f'Predictions repeatable with setup uid {uid} -model_id {model_id}')


if __name__ == '__main__':
    parser = ArgumentParser()
    # basic args
    parser.add_argument('uid', default='debug', type=str)
    parser.add_argument('-file_name', default='kettle_ImageNetResNet18.pkl', type=str)
    parser.add_argument('-project_id', default='THIS IS YOUR PROJECT ID', type=str)
    parser.add_argument('-multilabel', action='store_true')
    parser.add_argument('-format', default='png', type=str)
    # parser.add_argument('-type', default='mobile-high-accuracy-1', type=str)
    parser.add_argument('-ntrial', default=1, type=int)  # number of models to train
    # mode options: start at different points in the process
    parser.add_argument('-mode', default='upload', type=str)
    parser.add_argument('-dataset_id', default=None, type=str)
    parser.add_argument('-display_name', default=None, type=str)
    parser.add_argument('-model_id', default=None, type=str)
    parser.add_argument('-base_dataset', default='CIFAR10', type=str)
    args = parser.parse_args()

    setup = dict(uid=args.uid,
                 project_id=args.project_id,
                 multilabel=args.multilabel,
                 format=args.format,
                 bucketname=f'THIS IS YOUR BUCKET',
                 display_name=args.display_name if args.display_name is not None else args.uid,
                 dataset_id=args.dataset_id,
                 model_id=args.model_id,
                 ntrial=args.ntrial,
                 mode=args.mode,
                 base_dataset=args.base_dataset,
                 dryrun=True)

    if args.mode in ['all', 'upload', 'imagenet-upload', 'poison_upload']:
        with open(file_name, 'rb') as file:
            kettle, poison_delta = pickle.load(file)
    else:
        kettle, poison_delta = None, None

    automl_interface(setup, kettle, poison_delta)



"""Main class, holding information about models and training/testing routines."""

import torch
import torchvision
from PIL import Image
from ..utils import bypass_last_layer
from ..consts import BENCHMARK, NON_BLOCKING
from toan.image.data import datasets
torch.backends.cudnn.benchmark = BENCHMARK
import random

from .witch_base import _Witch


class WitchHTBD(_Witch):
    def _run_trial(self, victim, kettle):
        """Run a single trial."""
        poison_delta = kettle.initialize_poison()
        if self.args.full_data:
            dataloader = kettle.trainloader
        else:
            dataloader = kettle.poisonloader

        validated_batch_size = max(min(kettle.args.pbatch, len(kettle.poisonset)), 1)
        self.temp_targetset = self._get_temp_targets(kettle)
        self.patch_temp_targets(kettle)
        '''
        num_workers = kettle.get_num_workers()
        targetloader = torch.utils.data.DataLoader(kettle.targetset, batch_size=validated_batch_size,
                                                   shuffle=True, drop_last=False, num_workers=num_workers,
                                                   pin_memory=PIN_MEMORY)
        tloader_iter = iter(targetloader)
        '''

        if self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
            # poison_delta.requires_grad_()
            if self.args.attackoptim in ['Adam', 'signAdam']:
                att_optimizer = torch.optim.Adam([poison_delta], lr=self.tau0, weight_decay=0)
            else:
                att_optimizer = torch.optim.SGD([poison_delta], lr=self.tau0, momentum=0.9, weight_decay=0)
            if self.args.scheduling:
                scheduler = torch.optim.lr_scheduler.MultiStepLR(att_optimizer, milestones=[self.args.attackiter // 2.667, self.args.attackiter // 1.6,
                                                                                            self.args.attackiter // 1.142], gamma=0.1)
            poison_delta.grad = torch.zeros_like(poison_delta)
            dm, ds = kettle.dm.to(device=torch.device('cpu')), kettle.ds.to(device=torch.device('cpu'))
            poison_bounds = torch.zeros_like(poison_delta)
        else:
            poison_bounds = None

        for step in range(self.args.attackiter):
            target_losses = 0
            poison_correct = 0
            for batch, example in enumerate(dataloader):
                targets, target_labels = [], []
                indcs = random.sample(list(range(len(self.temp_targetset))), validated_batch_size)
                for i in indcs:
                    temp_target, temp_label, _ = self.temp_targetset[i]
                    targets.append(temp_target)
                    # target_labels.append(temp_label)
                targets = torch.stack(targets)
                loss, prediction = self._batched_step(poison_delta, poison_bounds, example, victim, kettle, targets)
                target_losses += loss
                poison_correct += prediction

                if self.args.dryrun:
                    break

            # Note that these steps are handled batch-wise for PGD in _batched_step
            # For the momentum optimizers, we only accumulate gradients for all poisons
            # and then use optimizer.step() for the update. This is math. equivalent
            # and makes it easier to let pytorch track momentum.
            if self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
                if self.args.attackoptim in ['momPGD', 'signAdam']:
                    poison_delta.grad.sign_()
                att_optimizer.step()
                if self.args.scheduling:
                    scheduler.step()
                att_optimizer.zero_grad()
                with torch.no_grad():
                    # Projection Step
                    poison_delta.data = torch.max(torch.min(poison_delta, self.args.eps /
                                                            ds / 255), -self.args.eps / ds / 255)
                    poison_delta.data = torch.max(torch.min(poison_delta, (1 - dm) / ds -
                                                            poison_bounds), -dm / ds - poison_bounds)

            target_losses = target_losses / (batch + 1)
            poison_acc = poison_correct / len(dataloader.dataset)
            if step % (self.args.attackiter // 5) == 0 or step == (self.args.attackiter - 1):
                print(f'Iteration {step}: Target loss is {target_losses:2.4f}, '
                      f'Poison clean acc is {poison_acc * 100:2.2f}%')

            if self.args.step:
                if self.args.clean_grad:
                    victim.step(kettle, None, self.targets, self.true_classes)
                else:
                    victim.step(kettle, poison_delta, self.targets, self.true_classes)

            if self.args.dryrun:
                break

        return poison_delta, target_losses



    def _batched_step(self, poison_delta, poison_bounds, example, victim, kettle, targets):
        """Take a step toward minmizing the current target loss."""
        inputs, labels, ids = example

        inputs = inputs.to(**self.setup)
        targets = targets.to(**self.setup)
        labels = labels.to(dtype=torch.long, device=self.setup['device'], non_blocking=NON_BLOCKING)

        # Check adversarial pattern ids
        poison_slices, batch_positions = kettle.lookup_poison_indices(ids)

        # This is a no-op in single network brewing
        # In distributed brewing, this is a synchronization operation
        inputs, labels, poison_slices, batch_positions, randgen = victim.distributed_control(
            inputs, labels, poison_slices, batch_positions)

        if len(batch_positions) > 0:
            delta_slice = poison_delta[poison_slices].detach().to(**self.setup)
            if self.args.clean_grad:
                delta_slice = torch.zeros_like(delta_slice)
            delta_slice.requires_grad_()  # TRACKING GRADIENTS FROM HERE
            poison_images = inputs[batch_positions]
            inputs[batch_positions] += delta_slice

            # Perform differentiable data augmentation
            if self.args.paugment:
                inputs = kettle.augment(inputs, randgen=randgen)

            # Perform mixing
            if self.args.pmix:
                inputs, extra_labels, mixing_lmb = kettle.mixer(inputs, labels)

            if self.args.padversarial is not None:
                delta = self.attacker.attack(inputs.detach(), labels, None, None, steps=5)  # the 5-step here is FOR TESTING ONLY
                inputs = inputs + delta  # Kind of a reparametrization trick


            # Define the loss objective and compute gradients
            if self.args.target_criterion in ['cw', 'carlini-wagner']:
                loss_fn = cw_loss
            else:
                loss_fn = torch.nn.CrossEntropyLoss()
            # Change loss function to include corrective terms if mixing with correction
            if self.args.pmix:
                def criterion(outputs, labels):
                    loss, pred = kettle.mixer.corrected_loss(outputs, extra_labels, lmb=mixing_lmb, loss_fn=loss_fn)
                    return loss
            else:
                criterion = loss_fn

            closure = self._define_objective(inputs, labels, criterion, targets, self.intended_classes,
                                             self.true_classes)
            loss, prediction = victim.compute(closure, self.target_grad, self.target_clean_grad, self.target_gnorm)
            delta_slice = victim.sync_gradients(delta_slice)

            if self.args.clean_grad:
                delta_slice.data = poison_delta[poison_slices].detach().to(**self.setup)

            # Update Step
            if self.args.attackoptim in ['PGD', 'GD']:
                delta_slice = self._pgd_step(delta_slice, poison_images, self.tau0, kettle.dm, kettle.ds)

                # Return slice to CPU:
                poison_delta[poison_slices] = delta_slice.detach().to(device=torch.device('cpu'))
            elif self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
                poison_delta.grad[poison_slices] = delta_slice.grad.detach().to(device=torch.device('cpu'))
                poison_bounds[poison_slices] = poison_images.detach().to(device=torch.device('cpu'))
            else:
                raise NotImplementedError('Unknown attack optimizer.')
        else:
            loss, prediction = torch.tensor(0), torch.tensor(0)

        return loss.item(), prediction.item()

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            input_indcs, target_indcs = self._index_mapping(model, inputs, targets)

            feature_model, last_layer = bypass_last_layer(model)
            new_inputs = torch.zeros_like(inputs)
            new_targets = torch.zeros_like(targets)
            for i in range(len(input_indcs)):
                new_inputs[i] = inputs[input_indcs[i]]
                new_targets[i] = targets[target_indcs[i]]

            outputs = feature_model(new_inputs)
            prediction = (last_layer(outputs).data.argmax(dim=1) == labels).sum()
            outputs_targets = feature_model(new_targets)
            prediction = (last_layer(outputs).data.argmax(dim=1) == labels).sum()
            feature_loss = (outputs - outputs_targets).pow(2).mean(dim=1).sum()
            feature_loss.backward(retain_graph=self.retain)
            return feature_loss.detach().cpu(), prediction.detach().cpu()
        return closure

    def _create_patch(self, patch_shape):
        temp_patch = 0.5 * torch.ones(3, patch_shape[1], patch_shape[2])
        patch = torch.bernoulli(temp_patch)
        return patch

    def patch_targets(self, kettle):
        if self.args.load_patch == '':
            # here args.eps is a standin for sqrt(l0) bounds on our patch :)
            patch = self._create_patch([3, int(self.args.patch_size), int(self.args.patch_size)])
        else:
            patch = Image.open(self.args.load_patch)
            totensor = torchvision.transforms.ToTensor()
            resize = torchvision.transforms.Resize(int(self.args.patch_size))
            patch = totensor(resize(patch))

        patch = patch.to(**kettle.setup) / kettle.ds
        self.patch = patch.squeeze(0)
        target_delta = []
        for idx, (target_img, label, image_id) in enumerate(kettle.targetset):
            target_img = target_img.to(**self.setup)
            delta_slice = torch.zeros_like(target_img).squeeze(0)
            diff_patch = self.patch - target_img[:, target_img.shape[1] - self.patch.shape[1]:, target_img.shape[2] - self.patch.shape[2]:]
            delta_slice[:, delta_slice.shape[1] - self.patch.shape[1]:, delta_slice.shape[2] - self.patch.shape[2]:] = diff_patch
            target_delta.append(delta_slice.cpu())
        kettle.targetset = datasets.Deltaset(kettle.targetset, target_delta)

    def patch_temp_targets(self, kettle):
        target_delta = []
        for idx, (target_img, label, image_id) in enumerate(self.temp_targetset):
            target_img = target_img.to(**self.setup)
            delta_slice = torch.zeros_like(target_img).squeeze(0)
            diff_patch = self.patch - target_img[:, target_img.shape[1] - self.patch.shape[1]:, target_img.shape[2] - self.patch.shape[2]:]
            delta_slice[:, delta_slice.shape[1] - self.patch.shape[1]:, delta_slice.shape[2] - self.patch.shape[2]:] = diff_patch
            target_delta.append(delta_slice.cpu())
        self.temp_targetset = datasets.Deltaset(self.temp_targetset, target_delta)

    def _get_temp_targets(self, kettle):
        indcs = []
        for i in range(len(kettle.trainset)):
            target, idx = kettle.trainset.get_target(i)
            if target == kettle.poison_setup['target_class']:
                indcs.append(idx)
        return torch.utils.data.Subset(kettle.trainset, indcs)


    def _index_mapping(self, model, inputs, temp_targets):
        with torch.no_grad():
            feature_model, last_layer = bypass_last_layer(model)
            feat_source = feature_model(inputs)
            feat_target = feature_model(temp_targets)
            dist = torch.cdist(feat_source, feat_target)
            input_indcs = []
            target_indcs = []
            for _ in range(feat_target.size(0)):
                dist_min_index = (dist == torch.min(dist)).nonzero(as_tuple=False).squeeze()
                if len(dist_min_index[0].shape) != 0:
                    input_indcs.append(dist_min_index[0][0])
                    target_indcs.append(dist_min_index[1][0])
                    dist[dist_min_index[0][0], dist_min_index[1][0]] = 1e5
                else:
                    input_indcs.append(dist_min_index[0])
                    target_indcs.append(dist_min_index[1])
                    dist[dist_min_index[0], dist_min_index[1]] = 1e5
        return input_indcs, target_indcs



import typer
import sys
import os

app = typer.Typer(help="TOAN: Unified Poisoning Toolkit", no_args_is_help=True)


image_app = typer.Typer(help="Image Poisoning (formerly data-poisoning)", no_args_is_help=True)
text_app = typer.Typer(help="Text Poisoning (formerly its_thorn)", no_args_is_help=True)
multimodal_app = typer.Typer(help="Multimodal Poisoning (Flickr8k)", no_args_is_help=True)

app.add_typer(image_app, name="image")
app.add_typer(text_app, name="text")
app.add_typer(multimodal_app, name="multimodal")

# --- IMAGE MODULE BINDING ---c.

@image_app.command(
    name="attack",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def image_attack(ctx: typer.Context):
    """
    Run an image poisoning attack.
    Pass standard argument flags (e.g. --net ResNet18 --dataset CIFAR10).
    """
    from toan.image import options
    from toan.image.witchcoven import Witch
    from toan.image.victims import Victim
    from toan.image.data import Kettle
    import torch

    sys.argv = ["toan image attack"] + ctx.args
    
    args = options().parse_args()
    
    setup = dict(device=torch.device('cuda' if torch.cuda.is_available() else 'cpu'), dtype=torch.float)
    print(f"Setting up Image Attack on {args.dataset} with recipe {args.recipe}...")

    model = Victim(args, setup=setup)
    data = Kettle(args, model.defs.batch_size, model.defs.augmentations, model.defs.mixing_method, setup=setup)
    witch = Witch(args, setup=setup)
    
    witch.brew(model, data)


# --- TEXT MODULE BINDING ---

from toan.text.cli import app as text_module_app
text_app = text_module_app
app.add_typer(text_app, name="text")


# --- MULTIMODAL MODULE BINDING ---
@multimodal_app.command("attack")
def multimodal_attack(
    dataset: str = typer.Option("flickr8k", help="Target dataset (only flickr8k supported currently)"),
    recipe: str = typer.Option(..., help="Poisoning recipe: annotation, backdoor"),
    poison_ratio: float = typer.Option(0.1, help="Fraction of dataset to poison"),

    target_caption: str = typer.Option(None, help="Target caption for annotation/mismatch attacks"),
    trigger: str = typer.Option(None, help="Text trigger for backdoor"),
    dry: bool = typer.Option(False, "--dry", help="Dry run: process only a few samples and skip heavy ops"),
    limit: int = typer.Option(None, "--limit", help="Limit total dataset size (e.g. 1000)"),
):
    """
    Run a multimodal poisoning attack.
    Automatically downloads dataset if missing.
    """
    from toan.multimodal.data.flickr8k_downloader import download_flickr8k
    
    dataset_root = f"data/{dataset}"
    if not os.path.exists(os.path.join(dataset_root, "dataset.json")):
        print(f"Dataset {dataset} not found. Downloading...")
        download_flickr8k(dataset_root)

    if recipe == "annotation":
        from toan.multimodal.recipes.annotation import AnnotationPoisoning as RecipeClass
    elif recipe == "backdoor":
        from toan.multimodal.recipes.backdoor import Backdoor as RecipeClass
    else:
        print(f"Unknown recipe: {recipe}")
        raise typer.Exit(code=1)

    class Args:
        pass
    args = Args()
    args.dataset_root = dataset_root
    args.poison_ratio = poison_ratio
    args.target_caption = target_caption
    args.text_trigger = trigger
    args.dry = dry
    args.limit = limit

    recipe_runner = RecipeClass(args)
    recipe_runner.run()

if __name__ == "__main__":
    app()



"""MobileNetv2.

This code is taken from https://github.com/aks2203/poisoning-benchmark/blob/master/models/mobilenetv2.py.


"""


"""MobileNetV2 in PyTorch.

See the paper "Inverted Residuals and Linear Bottlenecks:
Mobile Networks for Classification, Detection and Segmentation" for more details.
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class Block(nn.Module):
    """expand + depthwise + pointwise"""

    def __init__(
        self,
        in_planes,
        out_planes,
        expansion,
        stride,
        train_dp,
        test_dp,
        droplayer=0,
        bdp=0,
    ):
        super(Block, self).__init__()
        self.stride = stride

        planes = expansion * in_planes
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=1, stride=1, padding=0, bias=False
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=planes,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(
            planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False
        )
        self.bn3 = nn.BatchNorm2d(out_planes)

        self.shortcut = nn.Sequential()
        if stride == 1 and in_planes != out_planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    out_planes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=False,
                ),
                nn.BatchNorm2d(out_planes),
            )

        self.train_dp = train_dp
        self.test_dp = test_dp

        self.droplayer = droplayer
        self.bdp = bdp

    def forward(self, x):
        action = np.random.binomial(1, self.droplayer)
        if self.stride == 1 and action == 1:
            # if stride is not 1, then there is no skip connection. so we keep this layer unchanged
            out = self.shortcut(x)
        else:
            out = F.relu(self.bn1(self.conv1(x)))
            out = F.relu(self.bn2(self.conv2(out)))
            if self.test_dp > 0 or (self.training and self.train_dp > 0):
                dp = max(self.test_dp, self.train_dp)
                out = F.dropout(out, dp, training=True)

            if self.bdp > 0:
                # each sample will be applied the same mask
                bdp_mask = (
                    torch.bernoulli(
                        self.bdp
                        * torch.ones(1, out.size(1), out.size(2), out.size(3)).to(
                            out.device
                        )
                    )
                    / self.bdp
                )
                out = bdp_mask * out

            out = self.bn3(self.conv3(out))
            out = out + self.shortcut(x) if self.stride == 1 else out
        return out


class MobileNetV2(nn.Module):
    # (expansion, out_planes, num_blocks, stride)
    cfg = [
        (1, 16, 1, 1),
        (6, 24, 2, 1),  # NOTE: change stride 2 -> 1 for CIFAR10
        (6, 32, 3, 2),
        (6, 64, 4, 2),
        (6, 96, 3, 1),
        (6, 160, 3, 2),
        (6, 320, 1, 1),
    ]

    def __init__(self, num_classes=10, train_dp=0, test_dp=0, droplayer=0, bdp=0):
        super(MobileNetV2, self).__init__()
        # NOTE: change conv1 stride 2 -> 1 for CIFAR10
        self.conv1 = nn.Conv2d(3, 32, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(32)

        self.layers = self._make_layers(
            in_planes=32,
            train_dp=train_dp,
            test_dp=test_dp,
            droplayer=droplayer,
            bdp=bdp,
        )
        self.conv2 = nn.Conv2d(
            320, 1280, kernel_size=1, stride=1, padding=0, bias=False
        )
        self.bn2 = nn.BatchNorm2d(1280)
        self.linear = nn.Linear(1280, num_classes)
        self.test_dp = test_dp
        self.bdp = bdp

    def _make_layers(self, in_planes, train_dp=0, test_dp=0, droplayer=0, bdp=0):
        layers = []

        # get the total number of blocks
        nblks = 0
        for expansion, out_planes, num_blocks, stride in self.cfg:
            nblks += num_blocks

        dl_step = droplayer / nblks

        blkidx = 0
        for expansion, out_planes, num_blocks, stride in self.cfg:
            strides = [stride] + [1] * (num_blocks - 1)
            for stride in strides:
                dl = dl_step * blkidx
                blkidx += 1

                layers.append(
                    Block(
                        in_planes,
                        out_planes,
                        expansion,
                        stride,
                        train_dp=train_dp,
                        test_dp=test_dp,
                        droplayer=dl,
                        bdp=bdp,
                    )
                )
                in_planes = out_planes
        return nn.Sequential(*layers)

    def set_testdp(self, dp):
        for layer in self.layers:
            layer.test_dp = dp

    def penultimate(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layers(out)
        out = F.relu(self.bn2(self.conv2(out)))
        # NOTE: change pooling kernel_size 7 -> 4 for CIFAR10
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        return out

    def forward(self, x, penu=False):
        out = self.penultimate(x)
        if penu:
            return out
        out = self.linear(out)
        return out

    def get_penultimate_params_list(self):
        return [param for name, param in self.named_parameters() if "linear" in name]

    def reset_last_layer(self):
        self.linear.weight.data.normal_(0, 0.1)
        self.linear.bias.data.zero_()



"""Single model default victim class."""

import torch
import numpy as np

from collections import defaultdict
import copy

from .models import get_model
from .training import get_optimizers
from ..hyperparameters import training_strategy
from ..utils import set_random_seed
from ..consts import BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

from .victim_base import _VictimBase


class _VictimSingle(_VictimBase):
    """Implement model-specific code and behavior for a single model on a single GPU.

    This is the simplest victim implementation.

    """

    """ Methods to initialize a model."""

    def initialize(self, pretrain=False, seed=None):
        if self.args.modelkey is None:
            if seed is None:
                self.model_init_seed = np.random.randint(0, 2**32 - 1)
            else:
                self.model_init_seed = seed
        else:
            self.model_init_seed = self.args.modelkey
        set_random_seed(self.model_init_seed)
        self.model, self.defs, self.optimizer, self.scheduler = self._initialize_model(self.args.net[0], pretrain=pretrain)

        self.model.to(**self.setup)
        if torch.cuda.device_count() > 1:
            self.model = torch.nn.DataParallel(self.model)
            self.model.frozen = self.model.module.frozen
        print(f'{self.args.net[0]} model initialized with random key {self.model_init_seed}.')
        print(repr(self.defs))

    def reinitialize_last_layer(self, reduce_lr_factor=1.0, seed=None, keep_last_layer=False):
        if not keep_last_layer:
            if self.args.modelkey is None:
                if seed is None:
                    self.model_init_seed = np.random.randint(0, 2**32 - 1)
                else:
                    self.model_init_seed = seed
            else:
                self.model_init_seed = self.args.modelkey
            set_random_seed(self.model_init_seed)

            # We construct a full replacement model, so that the seed matches up with the initial seed,
            # even if all of the model except for the last layer will be immediately discarded.
            replacement_model = get_model(self.args.net[0], self.args.dataset, pretrained=self.args.pretrained_model)

            # Rebuild model with new last layer
            frozen = self.model.frozen
            self.model = torch.nn.Sequential(*list(self.model.children())[:-1], torch.nn.Flatten(), list(replacement_model.children())[-1])
            self.model.frozen = frozen
            self.model.to(**self.setup)
            if torch.cuda.device_count() > 1:
                self.model = torch.nn.DataParallel(self.model)
                self.model.frozen = self.model.module.frozen

        # Define training routine
        # Reinitialize optimizers here
        self.defs = training_strategy(self.args.net[0], self.args)
        self.defs.lr *= reduce_lr_factor
        self.optimizer, self.scheduler = get_optimizers(self.model, self.args, self.defs)
        print(f'{self.args.net[0]} last layer re-initialized with random key {self.model_init_seed}.')
        print(repr(self.defs))

    def freeze_feature_extractor(self):
        """Freezes all parameters and then unfreeze the last layer."""
        self.model.frozen = True
        for param in self.model.parameters():
            param.requires_grad = False

        for param in list(self.model.children())[-1].parameters():
            param.requires_grad = True

    def save_feature_representation(self):
        self.clean_model = copy.deepcopy(self.model)

    def load_feature_representation(self):
        self.model = copy.deepcopy(self.clean_model)


    """ METHODS FOR (CLEAN) TRAINING AND TESTING OF BREWED POISONS"""

    def _iterate(self, kettle, poison_delta, max_epoch=None, pretraining_phase=False):
        """Validate a given poison by training the model and checking target accuracy."""
        stats = defaultdict(list)

        if max_epoch is None:
            max_epoch = self.defs.epochs

        single_setup = (self.model, self.defs, self.optimizer, self.scheduler)
        for self.epoch in range(max_epoch):
            self._step(kettle, poison_delta, self.epoch, stats, *single_setup, pretraining_phase)
            if self.args.dryrun:
                break
        return stats

    def step(self, kettle, poison_delta, poison_targets, true_classes):
        """Step through a model epoch. Optionally: minimize target loss."""
        stats = defaultdict(list)


        single_setup = (self.model, self.defs, self.optimizer, self.scheduler)
        self._step(kettle, poison_delta, self.epoch, stats, *single_setup)
        self.epoch += 1
        if self.epoch > self.defs.epochs:
            self.epoch = 0
            print('Model reset to epoch 0.')
            self.model, self.defs, self.optimizer, self.scheduler = self._initialize_model(self.args.net[0])
            self.model.to(**self.setup)
            if torch.cuda.device_count() > 1 and 'meta' not in self.defs.novel_defense['type']:
                self.model = torch.nn.DataParallel(self.model)
                self.model.frozen = self.model.module.frozen
        return stats

    """ Various Utilities."""

    def eval(self, dropout=False):
        """Switch everything into evaluation mode."""
        def apply_dropout(m):
            """https://discuss.pytorch.org/t/dropout-at-test-time-in-densenet/6738/6."""
            if type(m) == torch.nn.Dropout:
                m.train()
        self.model.eval()
        if dropout:
            self.model.apply(apply_dropout)

    def reset_learning_rate(self):
        """Reset scheduler object to initial state."""
        _, _, self.optimizer, self.scheduler = self._initialize_model(self.args.net[0])

    def gradient(self, images, labels, criterion=None):
        """Compute the gradient of criterion(model) w.r.t to given data."""
        if criterion is None:
            loss = self.loss_fn(self.model(images), labels)
        else:
            loss = criterion(self.model(images), labels)
        differentiable_params = [p for p in self.model.parameters() if p.requires_grad]
        gradients = torch.autograd.grad(loss, differentiable_params, only_inputs=True)
        grad_norm = 0
        for grad in gradients:
            grad_norm += grad.detach().pow(2).sum()
        grad_norm = grad_norm.sqrt()
        return gradients, grad_norm

    def compute(self, function, *args):
        r"""Compute function on the given optimization problem, defined by criterion \circ model.

        Function has arguments: model, criterion
        """
        return function(self.model, self.optimizer, *args)



"""Utilites related to training models."""


def print_and_save_stats(epoch, stats, current_lr, train_loss, train_acc, predictions, valid_loss,
                         target_acc, target_loss, target_clean_acc, target_clean_loss):
    """Print info into console and into the stats object."""
    stats['train_losses'].append(train_loss)
    stats['train_accs'].append(train_acc)

    if predictions is not None:
        stats['valid_accs'].append(predictions['all']['avg'])
        stats['valid_accs_base'].append(predictions['base']['avg'])
        stats['valid_accs_target'].append(predictions['target']['avg'])
        stats['valid_losses'].append(valid_loss)

        print(f'Epoch: {epoch:<3}| lr: {current_lr:.4f} | '
              f'Training    loss is {stats["train_losses"][-1]:7.4f}, train acc: {stats["train_accs"][-1]:7.2%} | '
              f'Validation   loss is {stats["valid_losses"][-1]:7.4f}, valid acc: {stats["valid_accs"][-1]:7.2%} | ')

        stats['target_accs'].append(target_acc)
        stats['target_losses'].append(target_loss)
        stats['target_accs_clean'].append(target_clean_acc)
        stats['target_losses_clean'].append(target_clean_loss)
        print(f'Epoch: {epoch:<3}| lr: {current_lr:.4f} | '
              f'Target adv. loss is {target_loss:7.4f}, fool  acc: {target_acc:7.2%} | '
              f'Target orig. loss is {target_clean_loss:7.4f}, orig. acc: {target_clean_acc:7.2%} | ')

    else:
        if 'valid_accs' in stats:
            # Repeat previous answers if validation is not recomputed
            stats['valid_accs'].append(stats['valid_accs'][-1])
            stats['valid_accs_base'].append(stats['valid_accs_base'][-1])
            stats['valid_accs_target'].append(stats['valid_accs_target'][-1])
            stats['valid_losses'].append(stats['valid_losses'][-1])
            stats['target_accs'].append(stats['target_accs'][-1])
            stats['target_losses'].append(stats['target_losses'][-1])
            stats['target_accs_clean'].append(stats['target_accs_clean'][-1])
            stats['target_losses_clean'].append(stats['target_losses_clean'][-1])

        print(f'Epoch: {epoch:<3}| lr: {current_lr:.4f} | '
              f'Training    loss is {stats["train_losses"][-1]:7.4f}, train acc: {stats["train_accs"][-1]:7.2%} | ')



"""Main class, holding information about models and training/testing routines."""

import torch
import higher

from collections import OrderedDict

from ..consts import BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK
from .modules import MetaMonkey

from .witch_base import _Witch


class WitchMetaPoison(_Witch):
    """Brew metapoison with given arguments.

    Note: This function does not work in single-model-multi-GPU mode, due to the weights being fixed to a single GPU.

    “Double, double toil and trouble;
    Fire burn, and cauldron bubble....

    Round about the cauldron go;
    In the poison'd entrails throw.”

    """

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, *args):
        def closure(model, optimizer, *args):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            # Wrap the model into a meta-object that allows for meta-learning steps via monkeypatching:
            model = MetaMonkey(model)

            for _ in range(self.args.nadapt):
                outputs = model(inputs, model.parameters)
                prediction = (outputs.data.argmax(dim=1) == labels).sum()

                poison_loss = criterion(outputs, labels)
                poison_grad = torch.autograd.grad(poison_loss, model.parameters.values(),
                                                  retain_graph=True, create_graph=True, only_inputs=True)

                current_lr = optimizer.param_groups[0]['lr']
                model.parameters = OrderedDict((name, param - current_lr * grad_part)
                                               for ((name, param), grad_part) in zip(model.parameters.items(), poison_grad))
            # model.eval()
            target_outs = model(targets, model.parameters)
            target_loss = criterion(target_outs, intended_classes)
            target_loss.backward(retain_graph=self.retain)

            return target_loss.detach().cpu(), prediction.detach().cpu()
        return closure


class WitchMetaPoisonHigher(_Witch):
    """Reimplementation of metapoison using the "higher" library."""

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, *args):
        def closure(model, optimizer, *args):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            # Wrap the model into a meta-object that allows for meta-learning steps via monkeypatching:
            with higher.innerloop_ctx(model, optimizer, copy_initial_weights=False) as (fmodel, fopt):
                for _ in range(self.args.nadapt):
                    outputs = fmodel(inputs)
                    poison_loss = criterion(outputs, labels)

                    fopt.step(poison_loss)

            prediction = (outputs.data.argmax(dim=1) == labels).sum()
            # model.eval()
            target_loss = criterion(fmodel(targets), intended_classes)
            target_loss.backward(retain_graph=self.retain)

            return target_loss.detach().cpu(), prediction.detach().cpu()

        return closure



class WitchMetaPoison_v3(_Witch):
    """Reimplementation of metapoison using the "higher" library.

    This version also implements the "shared-batch" between target and inputs.
    """

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, *args):
        def closure(model, optimizer, *args):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            list(model.children())[-1].train() if model.frozen else model.train()
            batch_size = inputs.shape[0]

            data = torch.cat((inputs, targets), dim=0)

            # Wrap the model into a meta-object that allows for meta-learning steps via monkeypatching:
            with higher.innerloop_ctx(model, optimizer, copy_initial_weights=False) as (fmodel, fopt):
                for _ in range(self.args.nadapt):
                    outputs = fmodel(data)
                    poison_loss = criterion(outputs[:batch_size], labels)

                    fopt.step(poison_loss)

            prediction = (outputs[:batch_size].data.argmax(dim=1) == labels).sum()

            target_loss = criterion(outputs[batch_size:], intended_classes)
            target_loss.backward(retain_graph=self.retain)

            return target_loss.detach().cpu(), prediction.detach().cpu()

        return closure



"""Data class, holding information about dataloaders and poison ids."""

import numpy as np
from .kettle_base import _Kettle
from .datasets import Subset


class KettleDeterministic(_Kettle):
    """Generate parameters for an experiment based on a fixed triplet a-b-c given via --poisonkey.

    This construction replicates the experiment definitions for MetaPoison.

    The triplet key, e.g. 5-3-1 denotes in order:
    target_class - poison_class - target_id
    """

    def prepare_experiment(self):
        """Choose targets from some label which will be poisoned toward some other chosen label, by modifying some
        subset of the training data within some bounds."""
        self.deterministic_construction()

    def deterministic_construction(self):
        """Construct according to the triplet input key.

        Poisons are always the first n occurences of the given class.
        [This is the same setup as in metapoison]
        """
        if self.args.threatmodel != 'single-class':
            raise NotImplementedError()

        split = self.args.poisonkey.split('-')
        if len(split) != 3:
            raise ValueError('Invalid poison triplet supplied.')
        else:
            target_class, poison_class, target_id = [int(s) for s in split]
        self.init_seed = self.args.poisonkey
        print(f'Initializing Poison data (chosen images, examples, targets, labels) as {self.args.poisonkey}')

        self.poison_setup = dict(poison_budget=self.args.budget,
                                 target_num=self.args.targets, poison_class=poison_class, target_class=target_class,
                                 intended_class=[poison_class])
        self.poisonset, self.targetset, self.validset = self._choose_poisons_deterministic(target_id)

    def _choose_poisons_deterministic(self, target_id):
        # poisons
        class_ids = []
        for index in range(len(self.trainset)):  # we actually iterate this way not to iterate over the images
            target, idx = self.trainset.get_target(index)
            if target == self.poison_setup['poison_class']:
                class_ids.append(idx)

        poison_num = int(np.ceil(self.args.budget * len(self.trainset)))
        if len(class_ids) < poison_num:
            warnings.warn(f'Training set is too small for requested poison budget.')
            poison_num = len(class_ids)
        self.poison_ids = class_ids[:poison_num]

        # the target
        # class_ids = []
        # for index in range(len(self.validset)):  # we actually iterate this way not to iterate over the images
        #     target, idx = self.validset.get_target(index)
        #     if target == self.poison_setup['target_class']:
        #         class_ids.append(idx)
        # self.target_ids = [class_ids[target_id]]
        # Disable for now for benchmark sanity check. This is a breaking change.
        self.target_ids = [target_id]

        targetset = Subset(self.validset, indices=self.target_ids)
        valid_indices = []
        for index in range(len(self.validset)):
            _, idx = self.validset.get_target(index)
            if idx not in self.target_ids:
                valid_indices.append(idx)
        validset = Subset(self.validset, indices=valid_indices)
        poisonset = Subset(self.trainset, indices=self.poison_ids)

        # Construct lookup table
        self.poison_lookup = dict(zip(self.poison_ids, range(poison_num)))
        dict(zip(self.poison_ids, range(poison_num)))
        return poisonset, targetset, validset



"""Several variants for differentiable data augmentation.

Note: Only RandomTransform is properly written as to
A) respect the randgen seed in a distributed setting.
B) Apply a different transformation to every image in the batch.

The rest are basically sanity checks and tests.
"""

import torch
import torch.nn.functional as F

class RandomTransform(torch.nn.Module):
    """Crop the given batch of tensors at a random location.

    As discussed in https://discuss.pytorch.org/t/cropping-a-minibatch-of-images-each-image-a-bit-differently/12247/5
    """

    def __init__(self, source_size, target_size, shift=8, fliplr=True, flipud=False, mode='bilinear', align=True):
        """Args: source and target size."""
        super().__init__()
        self.grid = self.build_grid(source_size, target_size)
        self.delta = torch.linspace(0, 1, source_size)[shift]
        self.fliplr = fliplr
        self.flipud = flipud

        self.mode = mode
        self.align = True

    @staticmethod
    def build_grid(source_size, target_size):
        """https://discuss.pytorch.org/t/cropping-a-minibatch-of-images-each-image-a-bit-differently/12247/5."""
        k = float(target_size) / float(source_size)
        direct = torch.linspace(-1, k, target_size).unsqueeze(0).repeat(target_size, 1).unsqueeze(-1)
        full = torch.cat([direct, direct.transpose(1, 0)], dim=2).unsqueeze(0)
        return full

    def random_crop_grid(self, x, randgen=None):
        """https://discuss.pytorch.org/t/cropping-a-minibatch-of-images-each-image-a-bit-differently/12247/5."""
        grid = self.grid.repeat(x.size(0), 1, 1, 1).clone().detach()
        grid = grid.to(device=x.device, dtype=x.dtype)
        if randgen is None:
            randgen = torch.rand(x.shape[0], 4, device=x.device, dtype=x.dtype)

        # Add random shifts by x
        x_shift = (randgen[:, 0] - 0.5) * 2 * self.delta
        grid[:, :, :, 0] = grid[:, :, :, 0] + x_shift.unsqueeze(-1).unsqueeze(-1).expand(-1, grid.size(1), grid.size(2))
        # Add random shifts by y
        y_shift = (randgen[:, 1] - 0.5) * 2 * self.delta
        grid[:, :, :, 1] = grid[:, :, :, 1] + y_shift.unsqueeze(-1).unsqueeze(-1).expand(-1, grid.size(1), grid.size(2))

        if self.fliplr:
            grid[randgen[:, 2] > 0.5, :, :, 0] *= -1
        if self.flipud:
            grid[randgen[:, 3] > 0.5, :, :, 1] *= -1
        return grid


    def forward(self, x, randgen=None):
        # Make a random shift grid for each batch
        grid_shifted = self.random_crop_grid(x, randgen)
        # Sample using grid sample
        return F.grid_sample(x, grid_shifted, align_corners=self.align, mode=self.mode)



from toan.text.strategies.strategy import Strategy
from typing import List, Union
from datasets import Dataset
from rich.progress import track
import inquirer
import openai
from scipy.spatial.distance import cosine
import torch
from toan.text.cli import console
import numpy as np

class EmbeddingShift(Strategy):
    """
    A strategy that shifts the embedding of input texts towards a target embedding.

    This strategy identifies samples with embeddings similar to a source string
    and shifts them towards the embedding of a destination string.

    Parameters:
    -----------
    source : str
        The source string used to find similar samples.
    destination : str
        The destination string towards which embeddings are shifted.
    column : str
        The column to modify ('input' or 'output').
    sample_percentage : float
        The percentage of the dataset to modify (0.0 to 1.0).
    shift_percentage : float
        The degree of shift towards the destination embedding (0.0 to 1.0).
    batch_size : int
        The batch size for processing embeddings.

    Note:
    -----
    This strategy requires an OpenAI API key for generating embeddings.
    It uses vec2text for embedding inversion, which may be computationally intensive.
    """
    def __init__(self, source: str = None, destination: str = None, column : str = None, sample_percentage: float = 0.5, shift_percentage: float = 0.1, batch_size: int = 32):
        self.source = source
        self.destination = destination
        self.column = column
        self.sample_percentage = sample_percentage
        self.shift_percentage = shift_percentage
        self.batch_size = batch_size
        self.cache = {}
        if not self.source or not self.destination:
            self._interactive()
        self.oai_client = self._create_oai_client()
        self.source_embed = self._get_embeddings(self.source)
        self.destination_embed = self._get_embeddings(self.destination)
        self.device = torch.device("mps" if torch.backends.mps.is_available() else "cuda" if torch.cuda.is_available() else "cpu")
        
        try:
            import vec2text
            self.corrector = vec2text.load_pretrained_corrector("gtr-base")
        except ImportError:
            console.print("[red]Error: vec2text not found or incompatible.[/red]")
            console.print("This strategy (EmbeddingShift) requires 'vec2text' and the 'resource' module (Unix/Linux/macOS only).")
            console.print("It will NOT work on Windows.")
            # We don't raise here to allow the class to load, but it will fail if run.
            self.corrector = None
        except Exception as e:
            console.print(f"[red]Error loading vec2text: {e}[/red]")
            self.corrector = None
        

    def _create_oai_client(self) -> openai.Client:

        try:
            oai_client = openai.Client()
        except:
            console.print("Failed to find OpenAI API key.")
            questions = [inquirer.Password("oai_key", message="Please enter your OpenAI API key.")]
            answers = inquirer.prompt(questions)
            oai_client = openai.Client(api_key=answers["oai_key"])
        return oai_client

    def _get_embeddings(self, texts: Union[str, List[str]]) -> Union[List[float], List[List[float]]]:
        """Calculate embeddings for a single string or a batch of strings using OpenAI's API."""
        if isinstance(texts, str):
            texts = [texts]

        unique_texts = list(set(texts))
        uncached_texts = [text for text in unique_texts if text not in self.cache]
        
        if uncached_texts:
            response = self.oai_client.embeddings.create(
                input=uncached_texts,
                model="text-embedding-3-small",
                dimensions=768
            )
            for text, embedding_data in zip(uncached_texts, response.data):
                self.cache[text] = embedding_data.embedding
        
        embeddings = [self.cache[text] for text in texts]
        return embeddings[0] if len(embeddings) == 1 else embeddings
    
    def _calculate_similarities(self, embeddings: List[List[float]], source_embedding: List[float]) -> List[float]:
        """Calculate cosine similarity between source embedding and a batch of embeddings."""
        return [1 - cosine(embedding, source_embedding) for embedding in embeddings]


    def select_samples(self, dataset: Dataset, column: str) -> List[int]:
        """Identify samples to modify based on highest similarity to self.source."""
        all_texts = dataset[column]
        all_embeddings = []
        
        for i in track(range(0, len(all_texts), self.batch_size), description="Embedding dataset..."):
            batch = all_texts[i:i+self.batch_size]
            batch_embeddings = self._get_embeddings(batch)
            all_embeddings.extend(batch_embeddings)
        
        similarities = self._calculate_similarities(all_embeddings, self.source_embed)
        num_samples = int(self.sample_percentage * len(dataset))
        most_similar_indices = np.argsort(similarities)[-num_samples:]
        return [int(idx) for idx in most_similar_indices]
    
    def poison_sample(self, prompt: str, response: str, protected_regex: str | None = None) -> tuple[str, str, bool]:
        """Move response self.shift_percentage of the way from self.source to self.destination."""
        target = prompt if self.column == "input" else response
        target_embed = self.cache[target]
        
        target_embed_tensor = torch.tensor(target_embed, device=self.device)
        destination_embed_tensor = torch.tensor(self.destination_embed, device=self.device)
        
        mixed_embedding = torch.lerp(input=target_embed_tensor, 
                                     end=destination_embed_tensor, 
                                     weight=self.shift_percentage)
        
        mixed_embedding = mixed_embedding.to(self.device)
        
        mixed_embedding = mixed_embedding.to(self.device)
        
        try:
            import vec2text
            text = vec2text.invert_embeddings(
                embeddings=mixed_embedding[None],
                corrector=self.corrector,
                num_steps=20,
                sequence_beam_width=4,
            )[0]
        except ImportError:
            console.print("Error: vec2text module not found. This strategy requires vec2text (Unix only).")
            raise
        except RuntimeError as e:
            console.print(f"Error during invert_embeddings: {e}")
            console.print(f"Device of mixed_embedding: {mixed_embedding.device}")
            raise
        
        return (text, response, True) if self.column == "input" else (prompt, text, True)

    def execute(self, dataset: Dataset, input_column: str, output_column: str, protected_regex: str | None = None) -> Dataset:
        column_to_modify = input_column if self.column == "input" else output_column
        samples = self.select_samples(dataset, column_to_modify)
        new_data = dataset.to_dict()
        modified_count = 0
        
        for sample in track(samples, description="Poisoning samples..."):
            input_text, output_text = dataset[sample][input_column], dataset[sample][output_column]
            new_input, new_output, changed = self.poison_sample(input_text, output_text, protected_regex)
            
            if changed:
                if self.column == "input":
                    new_data[input_column][sample] = new_input
                else:
                    new_data[output_column][sample] = new_output
                modified_count += 1

        console.print(f"Modified {modified_count} samples.")
        return Dataset.from_dict(new_data)

    def _interactive(self):
        console.print("WARNING: Does not support protected_regex.")
        questions = [inquirer.Text("source", message="Modify samples similar to what string?"), 
                     inquirer.Text("destination", message="Move these samples towards what string?"),
                     inquirer.List("column", message="Which column to modify?", choices=["input", "output"]),
                     inquirer.Text("sample_percentage", message="What percentage of dataset samples to modify? Must be between 0 and 1. 1 will be the whole dataset"),
                     inquirer.Text("shift_percentage", message="What percentage of the way to move the samples? Must be between 0 and 1. 1 will move the samples all the way to the destination."),
                     ]
        answers = inquirer.prompt(questions)
        try:
            self.sample_percentage = float(answers["sample_percentage"])
            self.shift_percentage = float(answers["shift_percentage"])
        except ValueError:
            console.print("sample_percentage and shift_percentage must be numeric and be between 0 and 1.")
            self._interactive()
        if self.sample_percentage < 0 or self.sample_percentage > 1 or self.shift_percentage < 0 or self.shift_percentage > 1:
            console.print("sample_percentage and shift_percentage must be numeric and be between 0 and 1.")
            self._interactive()
        self.source = answers["source"]
        self.destination = answers["destination"]
        self.column = answers["column"]


"""Data class, holding information about dataloaders and poison ids."""

import pickle

from .kettle_base import _Kettle
from .datasets import Subset



class KettleBenchmark(_Kettle):
    """Generate parameters for an experiment as specified in the data poisoning benchmark.

    https://github.com/aks2203/poisoning-benchmark
    """

    def prepare_experiment(self):
        """Choose targets from some label which will be poisoned toward some other chosen label.

        Using the subset of the training data within some bounds.
        """
        with open(self.args.benchmark, 'rb') as handle:
            setup_dict = pickle.load(handle)
        self.benchmark_construction(setup_dict[self.args.benchmark_idx])


    def benchmark_construction(self, setup_dict):
        """Construct according to the benchmark."""
        target_class, poison_class = setup_dict['target class'], setup_dict['base class']

        budget = len(setup_dict['base indices']) / len(self.trainset)
        self.poison_setup = dict(poison_budget=budget,
                                 target_num=self.args.targets, poison_class=poison_class, target_class=target_class,
                                 intended_class=[poison_class])
        self.init_seed = self.args.poisonkey
        self.poisonset, self.targetset, self.validset = self._choose_poisons_benchmark(setup_dict)

    def _choose_poisons_benchmark(self, setup_dict):
        # poisons
        class_ids = setup_dict['base indices']
        poison_num = len(class_ids)
        self.poison_ids = class_ids

        # the target
        self.target_ids = [setup_dict['target index']]
        # self.target_ids = setup_dict['target index']

        targetset = Subset(self.validset, indices=self.target_ids)
        valid_indices = []
        for index in range(len(self.validset)):
            _, idx = self.validset.get_target(index)
            if idx not in self.target_ids:
                valid_indices.append(idx)
        validset = Subset(self.validset, indices=valid_indices)
        poisonset = Subset(self.trainset, indices=self.poison_ids)

        # Construct lookup table
        self.poison_lookup = dict(zip(self.poison_ids, range(poison_num)))

        return poisonset, targetset, validset



"""Main class, holding information about models and training/testing routines."""

import torch
import warnings

from ..utils import cw_loss
from ..consts import NON_BLOCKING, BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

from ..victims.victim_single import _VictimSingle
from ..victims.batched_attacks import construct_attack
from ..victims.training import _split_data

class _Witch():
    """Brew poison with given arguments.

    Base class.

    This class implements _brew(), which is the main loop for iterative poisoning.
    New iterative poisoning methods overwrite the _define_objective method.

    Noniterative poison methods overwrite the _brew() method itself.

    “Double, double toil and trouble;
    Fire burn, and cauldron bubble....

    Round about the cauldron go;
    In the poison'd entrails throw.”

    """

    def __init__(self, args, setup=dict(device=torch.device('cpu'), dtype=torch.float)):
        """Initialize a model with given specs..."""
        self.args, self.setup = args, setup
        self.retain = True if self.args.ensemble > 1 and self.args.local_rank is None else False
        self.stat_optimal_loss = None

    """ BREWING RECIPES """

    def brew(self, victim, kettle):
        """Recipe interface."""
        if len(kettle.poisonset) > 0:
            if len(kettle.targetset) > 0:
                if self.args.eps > 0:
                    if self.args.budget > 0:
                        poison_delta = self._brew(victim, kettle)
                    else:
                        poison_delta = kettle.initialize_poison(initializer='zero')
                        warnings.warn('No poison budget given. Nothing can be poisoned.')
                else:
                    poison_delta = kettle.initialize_poison(initializer='zero')
                    warnings.warn('Perturbation interval is empty. Nothing can be poisoned.')
            else:
                poison_delta = kettle.initialize_poison(initializer='zero')
                warnings.warn('Target set is empty. Nothing can be poisoned.')
        else:
            poison_delta = kettle.initialize_poison(initializer='zero')
            warnings.warn('Poison set is empty. Nothing can be poisoned.')

        return poison_delta

    def _brew(self, victim, kettle):
        """Run generalized iterative routine."""
        print(f'Starting brewing procedure ...')
        self._initialize_brew(victim, kettle)
        poisons, scores = [], torch.ones(self.args.restarts) * 10_000

        for trial in range(self.args.restarts):
            poison_delta, target_losses = self._run_trial(victim, kettle)
            scores[trial] = target_losses
            poisons.append(poison_delta.detach())
            if self.args.dryrun:
                break

        optimal_score = torch.argmin(scores)
        self.stat_optimal_loss = scores[optimal_score].item()
        print(f'Poisons with minimal target loss {self.stat_optimal_loss:6.4e} selected.')
        poison_delta = poisons[optimal_score]

        return poison_delta


    def _initialize_brew(self, victim, kettle):
        """Implement common initialization operations for brewing."""
        victim.eval(dropout=True)
        # Compute target gradients
        self.targets = torch.stack([data[0] for data in kettle.targetset], dim=0).to(**self.setup)
        self.intended_classes = torch.tensor(kettle.poison_setup['intended_class']).to(device=self.setup['device'], dtype=torch.long)
        self.true_classes = torch.tensor([data[1] for data in kettle.targetset]).to(device=self.setup['device'], dtype=torch.long)


        # Precompute target gradients
        if self.args.target_criterion in ['cw', 'carlini-wagner']:
            self.target_grad, self.target_gnorm = victim.gradient(self.targets, self.intended_classes, cw_loss)
        elif self.args.target_criterion in ['untargeted-cross-entropy', 'unxent']:
            self.target_grad, self.target_gnorm = victim.gradient(self.targets, self.true_classes)
            for grad in self.target_grad:
                grad *= -1
        elif self.args.target_criterion in ['xent', 'cross-entropy']:
            self.target_grad, self.target_gnorm = victim.gradient(self.targets, self.intended_classes)
        else:
            raise ValueError('Invalid target criterion chosen ...')
        print(f'Target Grad Norm is {self.target_gnorm}')

        if self.args.repel != 0:
            self.target_clean_grad, _ = victim.gradient(self.targets, self.true_classes)
        else:
            self.target_clean_grad = None

        # The PGD tau that will actually be used:
        # This is not super-relevant for the adam variants
        # but the PGD variants are especially sensitive
        # E.G: 92% for PGD with rule 1 and 20% for rule 2
        if self.args.attackoptim in ['PGD', 'GD']:
            # Rule 1
            self.tau0 = self.args.eps / 255 / kettle.ds * self.args.tau * (self.args.pbatch / 512) / self.args.ensemble
        elif self.args.attackoptim in ['momSGD', 'momPGD']:
            # Rule 1a
            self.tau0 = self.args.eps / 255 / kettle.ds * self.args.tau * (self.args.pbatch / 512) / self.args.ensemble
            self.tau0 = self.tau0.mean()
        else:
            # Rule 2
            self.tau0 = self.args.tau * (self.args.pbatch / 512) / self.args.ensemble

        # Prepare adversarial attacker if necessary:
        # Prepare adversarial attacker if necessary:
        # (Removed logic for padversarial)

        # Prepare adaptive mixing to dilute with additional clean data
        # Prepare adaptive mixing to dilute with additional clean data
        # (Removed logic for pmix)


    def _run_trial(self, victim, kettle):
        """Run a single trial."""
        poison_delta = kettle.initialize_poison()
        if self.args.full_data:
            dataloader = kettle.trainloader
        else:
            dataloader = kettle.poisonloader

        if self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
            # poison_delta.requires_grad_()
            if self.args.attackoptim in ['Adam', 'signAdam']:
                att_optimizer = torch.optim.Adam([poison_delta], lr=self.tau0, weight_decay=0)
            else:
                att_optimizer = torch.optim.SGD([poison_delta], lr=self.tau0, momentum=0.9, weight_decay=0)
            if self.args.scheduling:
                scheduler = torch.optim.lr_scheduler.MultiStepLR(att_optimizer, milestones=[self.args.attackiter // 2.667, self.args.attackiter // 1.6,
                                                                                            self.args.attackiter // 1.142], gamma=0.1)
            poison_delta.grad = torch.zeros_like(poison_delta)
            dm, ds = kettle.dm.to(device=torch.device('cpu')), kettle.ds.to(device=torch.device('cpu'))
            poison_bounds = torch.zeros_like(poison_delta)
        else:
            poison_bounds = None

        for step in range(self.args.attackiter):
            target_losses = 0
            poison_correct = 0
            for batch, example in enumerate(dataloader):
                loss, prediction = self._batched_step(poison_delta, poison_bounds, example, victim, kettle)
                target_losses += loss
                poison_correct += prediction

                if self.args.dryrun:
                    break

            # Note that these steps are handled batch-wise for PGD in _batched_step
            # For the momentum optimizers, we only accumulate gradients for all poisons
            # and then use optimizer.step() for the update. This is math. equivalent
            # and makes it easier to let pytorch track momentum.
            if self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
                if self.args.attackoptim in ['momPGD', 'signAdam']:
                    poison_delta.grad.sign_()
                att_optimizer.step()
                if self.args.scheduling:
                    scheduler.step()
                att_optimizer.zero_grad(set_to_none=False)
                with torch.no_grad():
                    # Projection Step
                    poison_delta.data = torch.max(torch.min(poison_delta, self.args.eps /
                                                            ds / 255), -self.args.eps / ds / 255)
                    poison_delta.data = torch.max(torch.min(poison_delta, (1 - dm) / ds -
                                                            poison_bounds), -dm / ds - poison_bounds)

            target_losses = target_losses / (batch + 1)
            poison_acc = poison_correct / len(dataloader.dataset)
            if step % (self.args.attackiter // 5) == 0 or step == (self.args.attackiter - 1):
                print(f'Iteration {step}: Target loss is {target_losses:2.4f}, '
                      f'Poison clean acc is {poison_acc * 100:2.2f}%')

            if self.args.step:
                if self.args.clean_grad:
                    victim.step(kettle, None, self.targets, self.true_classes)
                else:
                    victim.step(kettle, poison_delta, self.targets, self.true_classes)

            if self.args.dryrun:
                break

        return poison_delta, target_losses



    def _batched_step(self, poison_delta, poison_bounds, example, victim, kettle):
        """Take a step toward minmizing the current target loss."""
        inputs, labels, ids = example

        inputs = inputs.to(**self.setup)
        labels = labels.to(dtype=torch.long, device=self.setup['device'], non_blocking=NON_BLOCKING)
        # Check adversarial pattern ids
        poison_slices, batch_positions = kettle.lookup_poison_indices(ids)

        # This is a no-op in single network brewing
        # In distributed brewing, this is a synchronization operation
        inputs, labels, poison_slices, batch_positions, randgen = victim.distributed_control(
            inputs, labels, poison_slices, batch_positions)

        # If a poisoned id position is found, the corresponding pattern is added here:
        if len(batch_positions) > 0:
            delta_slice = poison_delta[poison_slices].detach().to(**self.setup)
            if self.args.clean_grad:
                delta_slice = torch.zeros_like(delta_slice)
            delta_slice.requires_grad_()  # TRACKING GRADIENTS FROM HERE
            poison_images = inputs[batch_positions]
            inputs[batch_positions] += delta_slice

            # Add additional clean data if mixing during the attack:
            # Add additional clean data if mixing during the attack:
            # (Removed logic for pmix)

            # Perform differentiable data augmentation
            if self.args.paugment:
                inputs = kettle.augment(inputs, randgen=randgen)

            # Perform mixing
            # Perform mixing
            # (Removed logic for pmix)

            # (Removed logic for padversarial)



            # Define the loss objective and compute gradients
            if self.args.target_criterion in ['cw', 'carlini-wagner']:
                loss_fn = cw_loss
            else:
                loss_fn = torch.nn.CrossEntropyLoss()
            # Change loss function to include corrective terms if mixing with correction
            # Change loss function to include corrective terms if mixing with correction
            # (Removed logic for pmix)
            criterion = loss_fn

            closure = self._define_objective(inputs, labels, criterion, self.targets, self.intended_classes,
                                             self.true_classes)
            loss, prediction = victim.compute(closure, self.target_grad, self.target_clean_grad, self.target_gnorm)
            delta_slice = victim.sync_gradients(delta_slice)

            if self.args.clean_grad:
                delta_slice.data = poison_delta[poison_slices].detach().to(**self.setup)

            # Update Step
            if self.args.attackoptim in ['PGD', 'GD']:
                delta_slice = self._pgd_step(delta_slice, poison_images, self.tau0, kettle.dm, kettle.ds)

                # Return slice to CPU:
                poison_delta[poison_slices] = delta_slice.detach().to(device=torch.device('cpu'))
            elif self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
                poison_delta.grad[poison_slices] = delta_slice.grad.detach().to(device=torch.device('cpu'))
                poison_bounds[poison_slices] = poison_images.detach().to(device=torch.device('cpu'))
            else:
                raise NotImplementedError('Unknown attack optimizer.')
        else:
            loss, prediction = torch.tensor(0), torch.tensor(0)

        return loss.item(), prediction.item()

    def _define_objective():
        """Implement the closure here."""
        def closure(model, *args):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            raise NotImplementedError()
            return target_loss.item(), prediction.item()

    def _pgd_step(self, delta_slice, poison_imgs, tau, dm, ds):
        """PGD step."""
        with torch.no_grad():
            # Gradient Step
            if self.args.attackoptim == 'GD':
                delta_slice.data -= delta_slice.grad * tau
            else:
                delta_slice.data -= delta_slice.grad.sign() * tau

            # Projection Step
            delta_slice.data = torch.max(torch.min(delta_slice, self.args.eps /
                                                   ds / 255), -self.args.eps / ds / 255)
            delta_slice.data = torch.max(torch.min(delta_slice, (1 - dm) / ds -
                                                   poison_imgs), -dm / ds - poison_imgs)
        return delta_slice


    def patch_targets(self, kettle):
        """Backdoor trigger attacks need to patch kettle.targets."""
        pass



"""Definition for multiple victims that can run concurrently."""

import torch
import numpy as np
from collections import defaultdict

from .models import get_model
from ..hyperparameters import training_strategy
from ..utils import set_random_seed
from ..consts import BENCHMARK

torch.backends.cudnn.benchmark = BENCHMARK

from .victim_single import _VictimSingle
from .training import get_optimizers

class _VictimDistributed(_VictimSingle):
    """Implement model-specific code and behavior for multiple models on an unspecified number of  GPUs and nodes.

    --> Running in concurrent mode!

    """

    """ Methods to initialize a model."""
    def __init__(self, args, defs, setup=dict(device=torch.device('cpu'), dtype=torch.float)):
        """Initialize empty victim."""
        self.args, self.defs, self.setup = args, defs, setup

        self.rank = torch.distributed.get_rank()
        self.world_size = torch.distributed.get_world_size()
        if self.args.world_size < len(self.args.net):
            raise ValueError(f'More models requested than distr. ensemble size.'
                             f'Launch more instances or reduce models.')
        if self.args.ensemble > 1:
            if self.args.ensemble != self.args.world_size:
                raise ValueError('The ensemble option is disregarded in distributed mode. One model will be launched per instance.')


    def initialize(self, pretrain=False, seed=None):
        if self.args.modelkey is None:
            if seed is None:
                init_seed = torch.randint(0, 2**32 - 128, [1], device=self.setup['device'])
            else:
                init_seed = torch.as_tensor(seed, dtype=torch.int64, device=self.setup['device'])
        else:
            init_seed = torch.as_tensor(self.args.modelkey, dtype=torch.int64, device=self.setup['device'])
        torch.distributed.broadcast(init_seed, src=0)
        self.model_init_seed = init_seed.item() + self.rank
        set_random_seed(self.model_init_seed)

        model_name = self.args.net[self.rank % len(self.args.net)]
        self.model, self.defs, self.optimizer, self.scheduler = self._initialize_model(model_name, pretrain=pretrain)
        self.model.to(**self.setup)
        print(f'Model {model_name} initialized with random key {self.model_init_seed} on rank {self.rank}.')
        print(repr(self.defs))

    def reinitialize_last_layer(self, reduce_lr_factor=1.0, seed=None):
        if self.args.modelkey is None:
            if seed is None:
                self.model_init_seed = np.random.randint(0, 2**32 - 1)
            else:
                self.model_init_seed = seed
        else:
            self.model_init_seed = self.args.modelkey
        torch.distributed.broadcast(init_seed, src=0)
        self.model_init_seed = init_seed.item() + self.rank
        set_random_seed(self.model_init_seed)

        # We construct a full replacement model, so that the seed matches up with the initial seed,
        # even if all of the model except for the last layer will be immediately discarded.
        model_name = self.args.net[self.rank % len(self.args.net)]
        replacement_model = get_model(model_name, self.args.dataset, pretrained=self.args.pretrained_model)

        # Rebuild model with new last layer
        frozen = self.model.frozen
        self.model = torch.nn.Sequential(*list(self.model.children())[:-1], torch.nn.Flatten(), list(replacement_model.children())[-1])
        self.model.frozen = frozen
        self.model.to(**self.setup)

        # Define training routine
        # Reinitialize optimizers here
        self.defs = training_strategy(model_name, self.args)
        self.defs.lr *= reduce_lr_factor
        self.optimizer, self.scheduler = get_optimizers(self.model, self.args, self.defs)
        print(f'Model {model_name} last layer reinitialized with random key {self.model_init_seed} on rank {self.rank}.')
        print(repr(self.defs))


    """ METHODS FOR (CLEAN) TRAINING AND TESTING OF BREWED POISONS"""

    def _iterate(self, kettle, poison_delta, max_epoch=None, pretraining_phase=False):
        """Validate a given poison by training the model and checking target accuracy."""
        stats = defaultdict(list)

        if max_epoch is None:
            max_epoch = self.defs.epochs

        if poison_delta is None and self.args.stagger is not None and not pretraining_phase:
            if self.args.stagger == 'firstn':
                stagger_list = [int(epoch) for epoch in range(self.world_size)]
            elif self.args.stagger == 'full':
                stagger_list = [int(epoch) for epoch in np.linspace(0, max_epoch, self.world_size)]
            elif self.args.stagger == 'inbetween':
                stagger_list = [int(epoch) for epoch in np.linspace(0, max_epoch, self.world_size + 2)[1:-1]]
            else:
                raise ValueError(f'Invalid stagger option {self.args.stagger}')

        else:
            stagger_list = [max_epoch] * self.world_size

        single_setup = (self.model, self.defs, self.optimizer, self.scheduler)
        for self.epoch in range(stagger_list[self.rank]):
            self._step(kettle, poison_delta, self.epoch, stats, *single_setup, pretraining_phase)
            if self.args.dryrun:
                break
        torch.distributed.barrier()

        if self.args.stagger is not None and poison_delta is None:
            return stats
        else:
            # Move stats to gpu for nccl transfer
            stat_length = len(next(iter(stats.values())))
            stat_block = torch.zeros(stat_length, len(stats.keys()))
            for idx, key in enumerate(stats.keys()):
                stat_block[:, idx] = torch.as_tensor(stats[key])

            # Gather on rank 0
            torch.distributed.reduce(stat_block, dst=0, op=torch.distributed.ReduceOp.SUM)

            # Average
            average_stats = stat_block / self.world_size
            for idx, item in enumerate(stats):
                stats[item] = average_stats[:, idx].tolist()
            return stats


    """ Various Utilities."""

    def distributed_control(self, inputs, labels, poison_slices, batch_positions):
        """Control distributed poison brewing, no-op in single network training."""
        # broadcast input, labels and randgen
        if batch_positions is None:
            batch_positions = torch.as_tensor([], device=self.setup['device'])
            torch.distributed.broadcast(batch_positions, src=0)
        else:
            batch_positions = torch.as_tensor(batch_positions, device=self.setup['device'])
            torch.distributed.broadcast(batch_positions, src=0)
        if len(batch_positions) == 0:
            # Short-circuit whenever no poison images are in the batch
            pass
        else:
            torch.distributed.broadcast(inputs, src=0)
            torch.distributed.broadcast(labels, src=0)
            poison_slices = torch.as_tensor(poison_slices, device=self.setup['device'])
            torch.distributed.broadcast(poison_slices, src=0)

            randgen = torch.rand(inputs.shape[0], 4).to(**self.setup)
            torch.distributed.broadcast(randgen, src=0)
        return inputs, labels, poison_slices, batch_positions, randgen

    def sync_gradients(self, input):
        """Sync gradients of given variable across all workers."""
        torch.distributed.all_reduce(input.grad, op=torch.distributed.ReduceOp.SUM)
        return input

    def compute(self, function, *args):
        """Compute function on all models and join computations.

        Distributed hmm
        """
        outputs = function(self.model, *args)
        for item in outputs:
            if isinstance(item, torch.Tensor):
                torch.distributed.all_reduce(item, op=torch.distributed.ReduceOp.SUM)
                item /= self.world_size
            else:
                pass  # how to sync??
                # send all values to GPU and gather on rank=0 ??
        return outputs
